import pytest
import threading
from unittest.mock import call

def test_domino_run_dev(setup_mlflow_tracking_server, mocker, mlflow, tracing, logging):
        """
        DominoRun will contain the ai system configuration loggged as parameters and the summary metrics for its
                evaluation traces which are attached to the run only,
                and a logged model with the ai system configuration,
                and a default summarization metrics are computed for the evaluation traces

                the sklearn autolog function will be called
        """
        # must import logging from the module instead of the package
        # so that mocker works
        create_external_model_spy = mocker.spy(mlflow, "create_external_model")
        exp = mlflow.set_experiment("test_domino_run")

        @tracing.add_tracing(name="add_numbers", autolog_frameworks=['sklearn'], evaluator=lambda span: { 'add_numbers': span.outputs })
        def add_numbers(x, y):
                return x + y

        run_id = None
        with logging.DominoRun() as run:
                run_id = run.info.run_id
                add_numbers(1, 1)
                add_numbers(2, 2)

        # verify logged model created only once
        assert create_external_model_spy.call_count == 1, "create external model should be called once"

        models = mlflow.search_logged_models(experiment_ids=[exp.experiment_id], output_format='list')

        assert len(models) == 1

        model = models[0]
        # verify ai system config added as configuration
        assert model.params['chat_assistant.max_tokens'] == '1500'
        assert model.params['chat_assistant.model'] == 'gpt-3.5-turbo'
        assert model.params['chat_assistant.temperature'] == '0.7'
        assert model.params['version'] == '1.0'

        # verify evaluation traces not logged to model
        ts = mlflow.search_traces(experiment_ids=[exp.experiment_id], model_id=model.model_id, return_type='list')
        assert len(ts) == 0, "traces should not be logged to model"

        run = mlflow.get_run(run_id)

        # verify run has ai system config logged to it as parameters
        assert run.data.params['chat_assistant.max_tokens'] == '1500'
        assert run.data.params['chat_assistant.model'] == 'gpt-3.5-turbo'
        assert run.data.params['chat_assistant.temperature'] == '0.7'
        assert run.data.params['version'] == '1.0'

        # verify run has summary metrics logged to it
        # average of outputs is 2 + 4/2 = 3
        assert run.data.metrics['mean_add_numbers'] == 3, "average of add_numbers should be 3"

def test_domino_run_dev_custom_aggregator(setup_mlflow_tracking_server, mlflow, tracing, logging):
        """
        DominoRun will contain custom summarizaiton metrics for eval traces
        """
        exp = mlflow.set_experiment("test_domino_run_custom_aggregator")

        @tracing.add_tracing(name="median", evaluator=lambda span: { 'median': span.outputs })
        def for_median(x):
                return x

        @tracing.add_tracing(name="mean", evaluator=lambda span: { 'mean': span.outputs })
        def for_mean(x):
                return x

        @tracing.add_tracing(name="stdev", evaluator=lambda span: { 'stdev': span.outputs })
        def for_stdev(x):
                return x

        @tracing.add_tracing(name="min", evaluator=lambda span: { 'min': span.outputs })
        def for_min(x):
                return x

        @tracing.add_tracing(name="max", evaluator=lambda span: { 'max': span.outputs })
        def for_max(x):
                return x

        summarization_metrics = [
                ('median', 'median'),
                ('mean', 'mean'),
                ('stdev', 'stdev'),
                ('min', 'min'),
                ('max', 'max')
        ]
        run_id = None
        with logging.DominoRun(custom_summary_metrics=summarization_metrics) as run:
                run_id = run.info.run_id
                for i in range(1, 6):
                        for_median(i)
                        for_mean(i)
                        for_stdev(i)
                        for_min(i)
                        for_max(i)

        run = mlflow.get_run(run_id)

        # verify run has summary metrics logged to it
        # mean of outputs is 2 + 4/2 = 3
        # median is 2, 2, 4 = 2
        assert run.data.metrics['median_median'] == 3
        assert run.data.metrics['mean_mean'] == 3
        assert run.data.metrics['stdev_stdev'] == 1.581
        assert run.data.metrics['min_min'] == 1
        assert run.data.metrics['max_max'] == 5

def test_domino_run_dev_bad_custom_aggregator(setup_mlflow_tracking_server, mlflow, tracing, logging):
        """
        DominoRun will fail if one of the aggregators is invalid
        """
        exp = mlflow.set_experiment("test_domino_run_dev_bad_custom_aggregator")

        summarization_metrics = [('max', 'sdf')]

        with pytest.raises(ValueError):
                logging.DominoRun(custom_summary_metrics=summarization_metrics)

def test_domino_run_configure_experiment_name(setup_mlflow_tracking_server, mlflow, logging, tracing):
        """
        if an experiment name is provided, the DominoRun will create a run in that experiment
        and log traces to it
        """
        mlflow.create_experiment("test_domino_run_configure_experiment_name")
        exp_id = mlflow.create_experiment("test_domino_run_configure_experiment_name_other")

        @tracing.add_tracing(name="unit")
        def unit(x):
                return x

        run_id = None
        with logging.DominoRun("test_domino_run_configure_experiment_name_other") as run:
                run_id = run.info.run_id
                unit(1)

        run = mlflow.get_run(run_id)

        traces = mlflow.search_traces(experiment_ids=[exp_id], filter_string=f"trace.name = 'unit'")

        assert run.info.experiment_id == exp_id, "run should belong to test_domino_run_configure_experiment_name_other"
        assert len(traces) == 1

def test_domino_run_extend_current_run(setup_mlflow_tracking_server, mlflow, logging, tracing):
        """
        if a run_id is provided, then the DominoRun with add traces to that run
        """
        mlflow.set_experiment("test_domino_run_extend_current_run")

        @tracing.add_tracing(name="unit", evaluator=lambda span: { 'unit': span.outputs })
        def unit(x):
                return x

        first_run_id = None
        second_run_id = None

        with logging.DominoRun() as run:
                first_run_id = run.info.run_id
                unit(1)

        with logging.DominoRun(run_id=first_run_id) as run:
                second_run_id = run.info.run_id
                unit(2)

        traces = mlflow.search_traces(experiment_ids=[run.info.experiment_id], filter_string=f"metadata.mlflow.sourceRun = '{first_run_id}'", return_type='list')

        assert first_run_id == second_run_id, "Both runs should have the same run_id"
        assert len(traces) == 2, "There should be two traces for unit"

        # each domino run should have an external model linked to it
        models = mlflow.search_logged_models(experiment_ids=[run.info.experiment_id], output_format='list')
        assert [m.source_run_id for m in models] == [first_run_id, first_run_id]

def test_domino_run_should_not_swallow_exceptions(setup_mlflow_tracking_server, mlflow, logging):
        """
        If the user's code raises an exception, the DominoRun should allow user code to catch it
        """
        mlflow.set_experiment("test_domino_run_should_not_swallow_exceptions")

        with pytest.raises(ZeroDivisionError):
            with logging.DominoRun() as run:
                1/0

def test_domino_run_parallelized_logic(setup_mlflow_tracking_server, mlflow, logging, tracing):
        """
        Logic run in threads should execute normally
        """
        mlflow.set_experiment("test_domino_run_parallelized_logic")

        @tracing.add_tracing(name="a")
        def a(num):
                return num

        @tracing.add_tracing(name="b")
        def b(num):
                return num

        with logging.DominoRun():
                t1 = threading.Thread(target=a, args=(10,))
                t2 = threading.Thread(target=b, args=(10,))

                t1.start()
                t2.start()

                t1.join()
                t2.join()

        traces_a = mlflow.search_traces(filter_string="trace.name = 'a'", return_type='list')
        traces_b = mlflow.search_traces(filter_string="trace.name = 'b'", return_type='list')
        def get_run_id(trace):
                return trace.info.trace_metadata.get('mlflow.sourceRun')

        assert len(traces_a) == 1, "There should be one trace for a"
        assert len(traces_b) == 1, "There should be one trace for b"
        assert get_run_id(traces_a[0]) == get_run_id(traces_b[0]), "The a and b traces should belong to the same run"

def test_domino_run_extend_concluded_run_manual_evals_mean_logged(setup_mlflow_tracking_server, mlflow, tracing, logging):
        """
        When extending a concluded run, manual log_evaluation calls inside the DominoRun block
        are summarized and the average metric is logged to the run.
        """
        mlflow.set_experiment("test_domino_run_extend_concluded_run_manual_evals_mean_logged")

        @tracing.add_tracing(name="add_numbers")
        def add_numbers(x, y):
                return x + y

        # Create and conclude a run with two traces (outputs: 2 and 4)
        with mlflow.start_run() as parent_run:
                concluded_run_id = parent_run.info.run_id
                add_numbers(1, 1)
                add_numbers(2, 2)

        # Extend the concluded run and log manual evaluations; DominoRun should log mean summary (3)
        with logging.DominoRun(run_id=concluded_run_id):
                traces_resp = tracing.search_traces(run_id=concluded_run_id, trace_name="add_numbers")
                for t in traces_resp.data:
                        # use the function output as the evaluation value
                        value = t.spans[0].outputs
                        logging.log_evaluation(
                                trace_id=t.id,
                                name="helpfulness",
                                value=value,
                        )

        run = mlflow.get_run(concluded_run_id)
        # average of 2 + 4 = 3
        assert run.data.metrics['mean_helpfulness'] == 3, "average of helpfulness should be 3"

def test_domino_run_extend_concluded_run_manual_evals_custom_aggregator_logged(setup_mlflow_tracking_server, mlflow, tracing, logging):
        """
        When extending a concluded run, with a custom aggregator, manual log_evaluation calls inside the
        DominoRun block are summarized using the custom aggregator and logged to the run.
        """
        mlflow.set_experiment("test_domino_run_extend_concluded_run_manual_evals_custom_aggregator_logged")

        @tracing.add_tracing(name="add_numbers")
        def add_numbers(x, y):
                return x + y

        # Create and conclude a run with two traces (outputs: 2 and 4)
        with mlflow.start_run() as parent_run:
                concluded_run_id = parent_run.info.run_id
                add_numbers(1, 1)
                add_numbers(2, 2)

        # Extend the concluded run and log manual evaluations; DominoRun should log custom summary (max -> 4)
        custom_summary_metrics = [("helpfulness", "max")]
        with logging.DominoRun(run_id=concluded_run_id, custom_summary_metrics=custom_summary_metrics):
                traces_resp = tracing.search_traces(run_id=concluded_run_id, trace_name="add_numbers")
                for t in traces_resp.data:
                        value = t.spans[0].outputs
                        logging.log_evaluation(
                                trace_id=t.id,
                                name="helpfulness",
                                value=value,
                        )

        run = mlflow.get_run(concluded_run_id)
        # max of 2 and 4 is 4
        assert run.data.metrics['max_helpfulness'] == 4, "max of helpfulness should be 4"

def test_domino_run_recomputes_existing_aggregations(setup_mlflow_tracking_server, mlflow, tracing, logging):
        """
        When a run already has aggregated metrics (e.g., max_<metric>), a subsequent DominoRun
        on the same run_id recomputes those aggregations in addition to defaults.
        """
        exp = mlflow.set_experiment("test_domino_run_recomputes_existing_aggregations")

        @tracing.add_tracing(name="agg", evaluator=lambda span: { 'agg': span.outputs })
        def agg_fn(x):
                return x

        run_id = None
        # First run computes both default mean and custom max aggregations
        with logging.DominoRun(custom_summary_metrics=[('agg', 'mean'), ('agg', 'max')]) as run:
                run_id = run.info.run_id
                agg_fn(1)
                agg_fn(3)


        run = mlflow.get_run(run_id)
        assert run.data.metrics['mean_agg'] == 2, 'mean should be 2'
        assert run.data.metrics['max_agg'] == 3, 'max should be 3'

        # Second run continues the same run and adds a new value; expects recomputed max (and mean)
        with logging.DominoRun(run_id=run_id) as run2:
                agg_fn(5)

        run = mlflow.get_run(run_id)
        assert run.data.metrics['max_agg'] == 5, 'max should be 5'
        assert run.data.metrics['mean_agg'] == 3, 'mean should be 3'



import os
import pytest
import time

from domino import Domino
from domino.http_request_manager import _HttpRequestManager

os.environ["DOMINO_MAX_RETRIES"] = "3"


def test_versioning(requests_mock, dummy_hostname):
    """validates domino version checking is correct"""

    # Mock a typical response from the jobs status API endpoint (GET)
    requests_mock.get(f"{dummy_hostname}/version", json={"version": "5.10.0"})

    dom = Domino(host=dummy_hostname, project="rand_user/rand_project", api_key="rand_api_key")

    dep_version = dom.deployment_version().get("version")
    assert dep_version == "5.10.0"
    assert dom.requires_at_least("5.3.0")
    with pytest.raises(Exception):
        dom.requires_at_least("5.11.0")

def test_request_session(test_auth_base):
     request_manager = _HttpRequestManager(auth=test_auth_base)
     start_time = time.time()
     try:
         response = request_manager.request_session.get(
             'https://localhost:9999' # ConnectionError
         )
     except Exception as ex:
         print('It failed :(', ex.__class__.__name__)
     else:
         print('It eventually worked', response.status_code)
     finally:
         end_time = time.time()
         total_time = end_time - start_time
         assert(total_time > 5) # actual value should be around 6.0210....



# do not import all endpoints into this module because that uses a lot of memory and stack frames
# if you need the ability to import all endpoints from this module, import them with
# from domino._impl.custommetrics.apis.path_to_api import path_to_api

import enum


class PathValues(str, enum.Enum):
    API_METRIC_ALERTS_V1 = "/api/metricAlerts/v1"
    API_METRIC_VALUES_V1 = "/api/metricValues/v1"
    API_METRIC_VALUES_V1_MODEL_MONITORING_ID_METRIC = "/api/metricValues/v1/{modelMonitoringId}/{metric}"



from datetime import datetime, timedelta
import pytest
import os
from unittest.mock import patch
from typing import Callable, Optional

from domino.aisystems._client import client
from domino.aisystems.tracing._util import build_ai_system_experiment_name
from .conftest import TEST_AI_SYSTEMS_ENV_VARS
from .test_util import reset_prod_tracing

def fixture_create_traces():
        pytest.importorskip("mlflow")
        import mlflow
        @mlflow.trace(name="test_add")
        def test_add(x, y):
                return x + y

        # create traces
        with mlflow.start_run():
                test_add(1, 2)

def _add_prod_tags(traces, ai_system_id: str, ai_system_version: str):
        for t in traces:
                client.set_trace_tag(t.info.trace_id, "mlflow.domino.app_id", ai_system_id)
                client.set_trace_tag(t.info.trace_id, "mlflow.domino.app_version_id", ai_system_version)

def create_span_at_time(name: str, inputs: int, hours_ago: int, experiment_id: str):
        pytest.importorskip("mlflow")
        import mlflow

        dt = datetime.now() - timedelta(hours=hours_ago)
        ns = int(dt.timestamp() * 1e9)
        span = mlflow.start_span_no_context(name=name, inputs=inputs, experiment_id=experiment_id, start_time_ns=ns)
        span.end()

def fixture_create_prod_traces(
        ai_system_id: str,
        ai_system_version: str,
        trace_name: str,
        tracing,
        hours_ago: Optional[int] = None, # also used as input value for span
):
        """Creates prod ai system traces with a specific trace name"""
        pytest.importorskip("mlflow")
        import mlflow

        reset_prod_tracing()

        @tracing.add_tracing(name=trace_name)
        def one(x):
                return x

        env_vars = TEST_AI_SYSTEMS_ENV_VARS | {"DOMINO_AI_SYSTEM_IS_PROD": "true", "DOMINO_APP_ID": ai_system_id }
        with patch.dict(os.environ, env_vars, clear=True):
                tracing.init_tracing()
                if hours_ago is not None:
                        experiment = mlflow.get_experiment_by_name(build_ai_system_experiment_name(ai_system_id))
                        create_span_at_time(trace_name, hours_ago, hours_ago, experiment.experiment_id)
                else:
                        one(1)

                exp_name = build_ai_system_experiment_name(ai_system_id)
                exp = mlflow.get_experiment_by_name(exp_name)

                ts = mlflow.search_traces(experiment_ids=[exp.experiment_id], filter_string=f"trace.name = '{trace_name}'", return_type='list')

                # add prod tags (would be done by Domino deployment)
                _add_prod_tags(ts, ai_system_id, ai_system_version)



# coding: utf-8

# flake8: noqa

"""
    Domino Public API

    Public API endpoints for Custom Metrics  # noqa: E501

    The version of the OpenAPI document: 5.3.0
    Generated by: https://openapi-generator.tech
"""

__version__ = "1.0.0"

# import ApiClient
from domino._impl.custommetrics.api_client import ApiClient

# import Configuration
from domino._impl.custommetrics.configuration import Configuration

# import exceptions
from domino._impl.custommetrics.exceptions import OpenApiException
from domino._impl.custommetrics.exceptions import ApiAttributeError
from domino._impl.custommetrics.exceptions import ApiTypeError
from domino._impl.custommetrics.exceptions import ApiValueError
from domino._impl.custommetrics.exceptions import ApiKeyError
from domino._impl.custommetrics.exceptions import ApiException



# coding: utf-8

"""
    Domino Public API

    Public API endpoints for Custom Metrics  # noqa: E501

    The version of the OpenAPI document: 5.3.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from domino._impl.custommetrics import schemas  # noqa: F401


class TargetRangeV1(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "condition",
        }
        
        class properties:
            
            
            class condition(
                schemas.EnumBase,
                schemas.StrSchema
            ):
            
            
                class MetaOapg:
                    enum_value_to_name = {
                        "lessThan": "LESS_THAN",
                        "lessThanEqual": "LESS_THAN_EQUAL",
                        "greaterThan": "GREATER_THAN",
                        "greaterThanEqual": "GREATER_THAN_EQUAL",
                        "between": "BETWEEN",
                    }
                
                @schemas.classproperty
                def LESS_THAN(cls):
                    return cls("lessThan")
                
                @schemas.classproperty
                def LESS_THAN_EQUAL(cls):
                    return cls("lessThanEqual")
                
                @schemas.classproperty
                def GREATER_THAN(cls):
                    return cls("greaterThan")
                
                @schemas.classproperty
                def GREATER_THAN_EQUAL(cls):
                    return cls("greaterThanEqual")
                
                @schemas.classproperty
                def BETWEEN(cls):
                    return cls("between")
            lowerLimit = schemas.NumberSchema
            upperLimit = schemas.NumberSchema
            __annotations__ = {
                "condition": condition,
                "lowerLimit": lowerLimit,
                "upperLimit": upperLimit,
            }
    
    condition: MetaOapg.properties.condition
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["condition"]) -> MetaOapg.properties.condition: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["lowerLimit"]) -> MetaOapg.properties.lowerLimit: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["upperLimit"]) -> MetaOapg.properties.upperLimit: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["condition", "lowerLimit", "upperLimit", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["condition"]) -> MetaOapg.properties.condition: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["lowerLimit"]) -> typing.Union[MetaOapg.properties.lowerLimit, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["upperLimit"]) -> typing.Union[MetaOapg.properties.upperLimit, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["condition", "lowerLimit", "upperLimit", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *args: typing.Union[dict, frozendict.frozendict, ],
        condition: typing.Union[MetaOapg.properties.condition, str, ],
        lowerLimit: typing.Union[MetaOapg.properties.lowerLimit, decimal.Decimal, int, float, schemas.Unset] = schemas.unset,
        upperLimit: typing.Union[MetaOapg.properties.upperLimit, decimal.Decimal, int, float, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'TargetRangeV1':
        return super().__new__(
            cls,
            *args,
            condition=condition,
            lowerLimit=lowerLimit,
            upperLimit=upperLimit,
            _configuration=_configuration,
            **kwargs,
        )



import os
import random

import pytest

from domino.helpers import domino_is_reachable
from unittest.mock import patch


@pytest.fixture
def random_seq():
    rand_val = random.randint(1000, 8888)
    rand_val2 = random.randint(1000, 8888)
    return f"-{rand_val}-{rand_val2}"


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_datasets_create_and_ids(default_domino_client, random_seq):
    dataset_name = "My-Integration-Test-Dataset" + random_seq
    dataset_desc = "A dataset for testing purposes."
    new_dataset = default_domino_client.datasets_create(
        dataset_name=dataset_name, dataset_description=dataset_desc
    )

    datasets_ids = default_domino_client.datasets_ids(default_domino_client.project_id)

    assert dataset_name == new_dataset["datasetName"]
    assert dataset_desc == new_dataset["description"]
    assert new_dataset["datasetId"] in datasets_ids


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_datasets_list(default_domino_client):
    datasets = default_domino_client.datasets_list()
    project_datasets = default_domino_client.datasets_list(
        default_domino_client.project_id
    )
    dataset_id = project_datasets[0]["datasetId"]

    assert datasets is not None
    assert project_datasets is not None
    assert dataset_id is not None


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_datasets_update_details_and_name(default_domino_client, random_seq):
    datasets_id = default_domino_client.datasets_ids(default_domino_client.project_id)[
        1
    ]

    new_datasets_name = "My-New-Integration-Test-Dataset" + random_seq
    new_datasets_description = "My New Integration Test Dataset Description"

    new_dataset = default_domino_client.datasets_update_details(
        datasets_id, dataset_name=new_datasets_name
    )
    newer_dataset = default_domino_client.datasets_update_details(
        datasets_id, dataset_description=new_datasets_description
    )

    datasets_names = default_domino_client.datasets_names(
        default_domino_client.project_id
    )

    assert new_datasets_name == new_dataset["datasetName"]
    assert new_datasets_name in datasets_names
    assert new_datasets_description == newer_dataset["description"]


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_datasets_details(default_domino_client, random_seq):
    new_dataset_name = "My-New-Integration-Test-Dataset-3" + random_seq
    new_dataset_description = "My New Integration Test Dataset Description 4"

    new_dataset = default_domino_client.datasets_create(
        dataset_name=new_dataset_name, dataset_description=new_dataset_description
    )

    new_dataset_id = new_dataset["datasetId"]

    dataset_details = default_domino_client.datasets_details(new_dataset_id)

    assert dataset_details is not None
    assert new_dataset_name == new_dataset["datasetName"]
    assert new_dataset_description == new_dataset["description"]
    assert "projectId" in dataset_details.keys()
    assert "created" in dataset_details.keys()


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_datasets_upload(default_domino_client):
    datasets_id = default_domino_client.datasets_ids(default_domino_client.project_id)[
        0
    ]
    assert "test_datasets.py" in os.listdir("tests")
    local_path_to_file = "tests/test_datasets.py"
    response = default_domino_client.datasets_upload_files(datasets_id, local_path_to_file)

    assert "test_datasets.py" in response


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_datasets_upload_with_sub_dir(default_domino_client):
    datasets_id = default_domino_client.datasets_ids(default_domino_client.project_id)[
        0
    ]
    assert "test_datasets.py" in os.listdir("tests")
    local_path_to_file = "tests/test_datasets.py"
    response = default_domino_client.datasets_upload_files(datasets_id, local_path_to_file,
                                                           target_relative_path="sub_d")
    assert "test_datasets.py" in response


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
@patch('os.path.exists')
def test_datasets_upload_mixed_slash_path(mock_exists, default_domino_client):
    # Simulating windows style-path for an existent file
    mock_exists.return_value = True
    datasets_id = default_domino_client.datasets_ids(default_domino_client.project_id)[
        0
    ]
    assert "back\\slash.txt" in os.listdir("tests/assets")
    local_path_to_file = "tests/assets/back\\slash.txt"
    response  = default_domino_client.datasets_upload_files(datasets_id, 
                                                            local_path_to_file)
    assert "back\\slash.txt" in response


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
@patch('os.path.exists')
def test_datasets_upload_windows_path(mock_exists, default_domino_client):
    # Simulating windows style-path for an existent file
    mock_exists.return_value = True
    datasets_id = default_domino_client.datasets_ids(default_domino_client.project_id)[
        0
    ]
    assert "test_datasets.py" in os.listdir("tests")
    windows_local_path_to_file = "tests\\test_datasets.py"
    response  = default_domino_client.datasets_upload_files(datasets_id, 
                                                            windows_local_path_to_file)
    assert "test_datasets.py" in response


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
@patch('os.path.exists')
def test_datasets_upload_with_sub_dir_windows_path(mock_exists, default_domino_client):
    # Simulating windows style-path for an existent file
    mock_exists.return_value = True
    datasets_id = default_domino_client.datasets_ids(default_domino_client.project_id)[
        0
    ]
    assert "test_datasets.py" in os.listdir("tests")
    windows_local_path_to_file = "tests\\test_datasets.py"
    response = default_domino_client.datasets_upload_files(datasets_id, 
                                                           windows_local_path_to_file,
                                                           target_relative_path="sub_d")

    assert "test_datasets.py" in response


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
@patch('os.path.exists')
def test_datasets_upload_directory_windows_path(mock_exists, default_domino_client):
    # Simulating windows style-path for an existent file
    mock_exists.return_value = True
    datasets_id = default_domino_client.datasets_ids(default_domino_client.project_id)[
        0
    ]
    assert os.path.isdir("tests/assets")
    windows_local_path_to_dir = "tests/assets"
    response  = default_domino_client.datasets_upload_files(datasets_id, 
                                                            windows_local_path_to_dir)
    assert "tests/assets" in response


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_datasets_upload_non_existing_file(default_domino_client):
    datasets_id = default_domino_client.datasets_ids(default_domino_client.project_id)[
        1
    ]
    local_path_to_file = "non_existing_file.py"
    try:
        default_domino_client.datasets_upload_files(datasets_id, local_path_to_file)
        assert False
    except ValueError:
        assert True


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_datasets_remove(default_domino_client):
    datasets_ids = default_domino_client.datasets_ids(default_domino_client.project_id)
    default_domino_client.datasets_remove(datasets_ids[1:])

    new_datasets_ids = default_domino_client.datasets_ids(
        default_domino_client.project_id
    )

    assert datasets_ids[-1] not in new_datasets_ids



import pytest
import uuid

from domino.domino_enums import BillingTagSettingMode, BudgetLabel
from domino.helpers import domino_is_reachable


def get_short_id() -> str:
    return str(uuid.uuid4())[:8]


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
@pytest.mark.parametrize("new_limit", [0.0006, 0.0037])
@pytest.mark.parametrize("budget_label", [BudgetLabel.PROJECT, BudgetLabel.ORGANIZATION, BudgetLabel.BILLINGTAG])
def test_budgets_defaults(budget_label, new_limit, default_domino_client):
    """
    Test get and update default budgets
    """
    budget_defaults = default_domino_client.budget_defaults_list()
    assert len(budget_defaults) == 3

    default_domino_client.budget_defaults_update(budget_label, new_limit)
    budget_defaults = default_domino_client.budget_defaults_list()
    for budget in budget_defaults:
        if budget["budgetLabel"] == budget_label.value:
            assert (budget["limit"] == new_limit)


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_budgets_overrides(default_domino_client):
    """
    Test creating, updating and deleting budget overrides
    """

    budget_ids = ["6626b3c64fa2ef89b5def375", "6626b3cf9106c3938a2c5f01", "6626b3d73473d4a99c2c642b"]
    budget_limit: float = 0.2048

    curr_overrides = default_domino_client.budget_overrides_list()
    overrides_count_start = len(curr_overrides)

    default_domino_client.budget_override_create(BudgetLabel.PROJECT, budget_ids[0], budget_limit)
    default_domino_client.budget_override_create(BudgetLabel.ORGANIZATION, budget_ids[1], budget_limit)
    default_domino_client.budget_override_create(BudgetLabel.BILLINGTAG, budget_ids[2], budget_limit)

    budget_overrides = default_domino_client.budget_overrides_list()

    assert len(budget_overrides) >= overrides_count_start
    for budget in budget_overrides:
        if budget["labelId"] in budget_ids:
            assert budget["limit"] == budget_limit

    new_limit = 0.1024
    default_domino_client.budget_override_update(BudgetLabel.BILLINGTAG, budget_ids[2], new_limit)

    updated_override = default_domino_client.budget_overrides_list()
    assert len(updated_override) >= 3
    for budget in updated_override:
        if budget["labelId"] == budget_ids[2]:
            assert budget["limit"] == new_limit

    for budget_id in budget_ids:
        default_domino_client.budget_override_delete(budget_id)

    overrides_reset = default_domino_client.budget_overrides_list()
    assert len(overrides_reset) == overrides_count_start


@pytest.mark.skipif(  #
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_budgets_alerts_settings(default_domino_client):  # TODO
    """
    Test creating a budget with current project, and other projects
    """
    alert_settings = default_domino_client.budget_alerts_settings()
    assert 'alertsEnabled' in alert_settings.keys()

    default_domino_client.budget_alerts_settings_update(alerts_enabled=False, notify_org_owner=False)
    update_setting_1 = default_domino_client.budget_alerts_settings()
    assert update_setting_1["alertsEnabled"] is False
    assert update_setting_1["notifyOrgOwner"] is False

    default_domino_client.budget_alerts_settings_update(alerts_enabled=True, notify_org_owner=True)
    update_setting_2 = default_domino_client.budget_alerts_settings()
    assert update_setting_2["alertsEnabled"] is True
    assert update_setting_2["notifyOrgOwner"] is True

    new_target_1 = {
        BudgetLabel.PROJECT.value: ["projectszzz@gmail.com"],
        BudgetLabel.ORGANIZATION.value: ["orgzyyy@gmail.com"],
    }

    default_domino_client.budget_alerts_targets_update(new_target_1)
    updated_settings = default_domino_client.budget_alerts_settings()
    updated_settings_t1 = updated_settings["alertTargets"]

    for new_t in updated_settings_t1:
        if new_t["label"] in new_target_1:
            assert new_t["emails"] == new_target_1[new_t["label"]]

    new_target_2 = {
        BudgetLabel.PROJECT.value: ["projectsffff@gmail.com"],
        BudgetLabel.BILLINGTAG.value: ["billing-tag@gmail.com"],
    }

    default_domino_client.budget_alerts_targets_update(new_target_2)
    updated_settings = default_domino_client.budget_alerts_settings()
    updated_settings_t2 = updated_settings["alertTargets"]

    for new_t in updated_settings_t2:
        if new_t["label"] in new_target_2:
            assert new_t["emails"] == new_target_2[new_t["label"]]


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_billing_tag_settings(default_domino_client):
    """
    Test creating a budget with current project, and other projects
    """
    billing_tag_setting = default_domino_client.billing_tag_settings()
    assert 'mode' in billing_tag_setting.keys()

    default_domino_client.billing_tag_settings_mode_update(BillingTagSettingMode.REQUIRED)
    updated_tag_setting = default_domino_client.billing_tag_settings()
    assert updated_tag_setting["mode"] == BillingTagSettingMode.REQUIRED.value
    mode = default_domino_client.billing_tag_settings_mode()
    assert mode["mode"] == BillingTagSettingMode.REQUIRED.value

    default_domino_client.billing_tag_settings_mode_update(BillingTagSettingMode.OPTIONAL)
    newer_tag_setting = default_domino_client.billing_tag_settings()
    assert newer_tag_setting["mode"] == BillingTagSettingMode.OPTIONAL.value


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_billing_tags(default_domino_client):
    """
    Test creating a budget with current project, and other projects
    """
    billing_tags = ["PYTHON-DOMINO-ACTIVE-tag-001", "PYTHON-DOMINO-ACTIVE-tag-002", "PYTHON-DOMINO-ACTIVE-tag-003"]
    new_billing_tags = default_domino_client.billing_tags_create(billing_tags)
    assert len(new_billing_tags["billingTags"]) == 3

    active_tags = default_domino_client.billing_tags_list_active()
    assert len(active_tags["billingTags"]) >= 3

    active_tag = default_domino_client.active_billing_tag_by_name(billing_tags[0])
    assert active_tag["billingTag"]["tag"] == billing_tags[0]
    assert active_tag["billingTag"]["active"] is True

    default_domino_client.billing_tag_archive(billing_tags[0])
    archive_tag = default_domino_client.active_billing_tag_by_name(billing_tags[0])
    assert archive_tag["billingTag"]["tag"] == billing_tags[0]
    assert archive_tag["billingTag"]["active"] is False


@pytest.mark.skipif(
    not domino_is_reachable(), reason="No access to a live Domino deployment"
)
def test_projects_billing_tag(default_domino_client):
    """
    Test creating a budget with current project, and other projects
    """

    setting_mode = default_domino_client.billing_tag_settings_mode()
    if setting_mode["mode"] != BillingTagSettingMode.OPTIONAL.value:
        default_domino_client.billing_tag_settings_mode_update(BillingTagSettingMode.OPTIONAL)

    test_billing_tag = f"TestBillingTag-{get_short_id()}"
    test_billing_tag2 = f"TestBillingTag-{get_short_id()}"
    test_billing_tag3 = f"TestBillingTag-{get_short_id()}"
    default_domino_client.billing_tags_create([test_billing_tag, test_billing_tag2, test_billing_tag3])

    test_project_name = f"project-{get_short_id()}"
    test_project_name_2 = f"project-{get_short_id()}"

    bt_project = default_domino_client.project_create_v4(project_name=test_project_name, billing_tag=test_billing_tag)
    project = default_domino_client.project_create_v4(project_name=test_project_name_2)

    project_bt = default_domino_client.project_billing_tag(bt_project["id"])
    assert project_bt["tag"] == test_billing_tag

    project_bt_none = default_domino_client.project_billing_tag(project["id"])
    assert project_bt_none is None

    default_domino_client.project_billing_tag_update(test_billing_tag2, project["id"])
    project_update = default_domino_client.project_billing_tag(project["id"])
    assert project_update["tag"] == test_billing_tag2

    default_domino_client.project_billing_tag_reset(project["id"])
    project_bt_reset = default_domino_client.project_billing_tag(project["id"])
    assert project_bt_reset is None

    query_p = default_domino_client.projects_by_billing_tag(billing_tag=test_billing_tag)
    assert query_p["totalMatches"] == 1
    assert query_p["page"][0]["id"] == bt_project["id"]
    assert query_p["page"][0]["billingTag"]["tag"] == test_billing_tag

    projects_tags = {bt_project["id"]: test_billing_tag3, project["id"]: test_billing_tag3}
    default_domino_client.project_billing_tag_bulk_update(projects_tags)

    query_p = default_domino_client.projects_by_billing_tag(billing_tag=test_billing_tag3)
    assert query_p["totalMatches"] == 2
    assert query_p["page"][0]["id"] in {bt_project["id"], project["id"]}
    assert query_p["page"][0]["billingTag"]["tag"] == test_billing_tag3



__all__ = ["DominoOperator", "DominoSparkOperator"]

try:
    from domino.airflow._operator import DominoOperator, DominoSparkOperator
except SyntaxError:
    raise ImportError(
        "Use of the Airflow DominoOperator requires typing (Python 3.5+)."
    )



from enum import Enum


class BudgetLabel(Enum):
    BILLINGTAG = "BillingTag"
    ORGANIZATION = "Organization"
    PROJECT = "Project"


class BudgetType(Enum):
    OVERRIDE = "Override"
    DEFAULT = "Default"


class BillingTagSettingMode(Enum):
    DISABLED = "Disabled"
    OPTIONAL = "Optional"
    REQUIRED = "Required"


class ProjectVisibility(Enum):
    PRIVATE = "Private"
    PUBLIC = "Public"
    SEARCHABLE = "Searchable"



import os

import pytest


@pytest.mark.skipif(os.getenv("DATA_SDK") != "yes", reason="Extra dependency required")
def test_data_sources_import():
    from domino.data_sources import ADLSConfig, DataSourceClient

    assert DataSourceClient
    assert ADLSConfig


@pytest.mark.skipif(os.getenv("DATA_SDK") != "yes", reason="Extra dependency required")
def test_data_sources_import_cant_find():  # noqa
    try:
        from domino.data_sources import NotADatasourceConfig

        assert NotADatasourceConfig
    except ImportError as exc:
        assert (
            "cannot import name 'NotADatasourceConfig' from 'domino.data_sources'"
            in str(exc)
        )



import itertools
import logging
import mlflow
import re
from statistics import median, stdev
import traceback
from typing import Literal, Optional, Callable

from .._client import client
from .._constants import LARGEST_MAX_RESULTS_PAGE_SIZE, DOMINO_INTERNAL_EVAL_TAG
from .._eval_tags import build_metric_tag, VALID_LABEL_PATTERN
from .._verify_domino_support import verify_domino_support
from ..read_ai_system_config import get_flattened_ai_system_config

TOTAL_DECIMAL_PLACES = 3
DOMINO_EVAL_METRIC_TAG_PATTERN = f"domino.prog.metric.({VALID_LABEL_PATTERN})"
AGGREGATED_METRIC_PATTERN = r"^(mean|median|stdev|max|min)_(.+)$"

SummaryStatistic = Literal["mean", "median", "stdev", "max", "min"]


def get_all_traces_for_run(experiment_id: str, run_id: str):
    filter_string = f"metadata.mlflow.sourceRun = '{run_id}' AND tags.{DOMINO_INTERNAL_EVAL_TAG} = 'true'"

    logger.debug(f"Searching for traces with filter: {filter_string}")

    next_page_token = None

    traces = client.search_traces(
        experiment_ids=[experiment_id],
        filter_string=filter_string,
        page_token=next_page_token,
        max_results=LARGEST_MAX_RESULTS_PAGE_SIZE,
        order_by=["attributes.timestamp_ms ASC"],
    )
    next_page_token = traces.token
    while next_page_token is not None:
        next_traces = client.search_traces(
            experiment_ids=[experiment_id],
            filter_string=filter_string,
            page_token=next_page_token,
            max_results=LARGEST_MAX_RESULTS_PAGE_SIZE,
            order_by=["attributes.timestamp_ms ASC"],
        )

        next_page_token = next_traces.token
        traces.extend(next_traces)

    return traces


def is_metric_tag(tag: str) -> bool:
    return re.match(DOMINO_EVAL_METRIC_TAG_PATTERN, tag) is not None


def get_metric_tag_name(tag: str) -> Optional[str]:
    match = re.match(DOMINO_EVAL_METRIC_TAG_PATTERN, tag)
    if match:
        return match.group(1)
    return None


def _parse_aggregated_metric_key(metric_key: str) -> Optional[tuple[str, str]]:
    """
    Parse an aggregated metric key of the form <agg>_<metric_name>.
    Returns (agg, tag) or None
    """
    m = re.match(AGGREGATED_METRIC_PATTERN, metric_key)
    if not m:
        return None

    try:
        agg = m.group(1)
        tag = m.group(2)
        return agg, tag
    except Exception as e:
        logger.warning(f"Failed to parse aggregated metric key {metric_key}: {e}")
        return None


def _parse_value(v):
    try:
        return float(v)
    except ValueError as e:
        logger.warning(f"Failed to parse value {v} as float: {e}")
        raise e


def _get_experiment_id_from_name(name: Optional[str]):
    if name:
        exp = mlflow.get_experiment_by_name(name)
        if not exp:
            raise Exception(f"Could not find experiment with name {name}")
        return exp.experiment_id
    return None


def _choose_summarizer(statistic: SummaryStatistic) -> Callable[[list[float]], float]:
    match statistic:
        case "mean":
            return lambda values: round(sum(values) / len(values), TOTAL_DECIMAL_PLACES)
        case "median":
            return lambda values: median(sorted(values))
        case "stdev":
            return lambda values: round(stdev(values), TOTAL_DECIMAL_PLACES)
        case "max":
            return max
        case "min":
            return min
        case _:
            raise ValueError(f"Unknown summary statistic: {statistic}")

class DominoRun:
    def __init__(
        self,
        experiment_name: Optional[str] = None,
        run_id: Optional[str] = None,
        ai_system_config_path: Optional[str] = None,
        custom_summary_metrics: Optional[list[(str, SummaryStatistic)]] = None,
    ):
        """DominoRun is a context manager that starts an Mlflow run and attaches the user's AI System configuration to it,
        create a Logged Model with the AI System configuration, and computes summary metrics for evaluation traces made during the run.
        Average metrics are computed by default, but the user can provide a custom list of evaluation metric aggregators.
        This is intended to be used in development mode for AI System evaluation.
        Context manager docs: https://docs.python.org/3/library/contextlib.html

        Parallelism: DominoRun is not thread-safe. Runs in different threads will work correctly. This is due to
        Mlflow's architecture. Parallelizing operations within a single DominoRun context however, is supported.

        Example:
            import mlflow

            mlflow.set_experiment("my_experiment")

            with DominoRun():
                train_model()

        Args:
                experiment_name: the name of the mlflow experiment to log the run to.

                run_id: optional, the ID of the mlflow run to continue logging to. If not provided a new run will start.

                ai_system_config_path: the optional path to the AI System configuration file. If not provided, defaults to the DOMINO_AI_SYSTEM_CONFIG_PATH environment variable.

                custom_summary_metrics: an optional list of tuples that define what summary statistic to use with what evaluation metric. Valid summary statistics are: "mean", "median", "stdev", "max", "min" e.g. [("hallucination_rate", "max")]

        Returns: DominoRun context manager
        """
        verify_domino_support()
        self.experiment_name = experiment_name
        self.run_id = run_id
        self.config_path = ai_system_config_path
        self.custom_summary_metrics = custom_summary_metrics
        self.experiment_id = _get_experiment_id_from_name(experiment_name)
        self._run = None

        self._validate()

    def _validate(self):
        # validate custome summary metrics
        if self.custom_summary_metrics:
            for _, stat in self.custom_summary_metrics:
                _choose_summarizer(stat)

    def _start_run(self):
        """
        Starts an mlflow run, either by continueing an existing run or starting a new one,
        and saves the experiment ID of the run in order to support starting runs
        without a user specifying the experiment explicitly.
        """
        if not self._run:
            if self.run_id:
                self._run = mlflow.get_run(run_id=self.run_id)
                experiment_id = self._run.info.experiment_id
                self._run = mlflow.start_run(
                    experiment_id=experiment_id, run_id=self.run_id
                )

            elif self.experiment_id:
                # start the run in this experiment
                self._run = mlflow.start_run(experiment_id=self.experiment_id)
            else:
                self._run = mlflow.start_run()

            self.experiment_id = self._run.info.experiment_id

    def __enter__(self):
        """
        Called when the 'with' block is executed
        """
        self._start_run()
        self.__log_params(self._run.info.run_id)

        return self._run

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Called when the 'with' block ends
        """
        try:
            traces = get_all_traces_for_run(self.experiment_id, self._run.info.run_id)

            # group by evaluation name
            eval_tags = sorted(
                [
                    (key, value)
                    for t in traces
                    for (key, value) in t.info.tags.items()
                    if is_metric_tag(key)
                ],
                key=lambda x: x[0],
            )
            grouped_eval_tags = dict(
                [
                    (k, list(vs))
                    for (k, vs) in itertools.groupby(eval_tags, key=lambda x: x[0])
                ]
            )

            custom_summary_metrics = None
            if self.custom_summary_metrics:
                custom_summary_metrics = [
                    (build_metric_tag(tag), stat)
                    for tag, stat in self.custom_summary_metrics
                ]

            # Base specs: either user-provided or default to mean for discovered tags
            base_summary_metric_specs = custom_summary_metrics or [
                (tag, "mean") for tag, _ in grouped_eval_tags.items()
            ]

            # Find previously calculated aggregated metrics
            recompute_specs = []
            try:
                existing = mlflow.get_run(self._run.info.run_id)
                metric_keys = list(existing.data.metrics.keys()) if existing else []
                for k in metric_keys:
                    parsed = _parse_aggregated_metric_key(k)
                    if parsed:
                        agg, tag = parsed
                        recompute_specs.append((build_metric_tag(tag), agg))
            except Exception as e:
                logger.warning(
                    f"Unable to inspect existing aggregated metrics for recomputation: {e}"
                )

            # Combine and deduplicate
            summary_metric_specs = list(
                set(base_summary_metric_specs + recompute_specs)
            )

            for tag, summary_statistic in summary_metric_specs:
                aggregator = _choose_summarizer(summary_statistic)
                found_values = grouped_eval_tags.get(tag, None)
                if found_values:
                    try:
                        values = [_parse_value(v) for (_, v) in found_values]
                        summary = aggregator(values)
                        # Derive metric name from tag; fall back to tag with warning if parsing fails
                        metric_name = get_metric_tag_name(tag)
                        if metric_name is None:
                            logger.warning(
                                f"Could not extract metric name from tag '{tag}'. "
                                f"Falling back to using the full tag in aggregated metric name."
                            )
                            metric_name = tag

                        aggregated_metric_name = f"{summary_statistic}_{metric_name}"
                        mlflow.log_metric(
                            aggregated_metric_name,
                            summary,
                            run_id=self._run.info.run_id,
                        )
                    except Exception as e:
                        logger.error(
                            f"Failed to log summarization metric for {tag}: {e}"
                        )
                else:
                    logger.debug(
                        f"No evaluation tags found for {tag}, skipping summarization metric logging"
                    )
        except Exception as e:
            logger.error(
                f"Something went wrong while computing summarization metric for run {self._run.info.run_id}: {e}"
            )
            traceback.print_exc()
        finally:
            mlflow.end_run()

    def __log_params(self, run_id: str):
        """
        Saves the user's AI System configuration as parameters to the mlflow run and
        the AI System's logged model
        """
        try:
            params = get_flattened_ai_system_config(self.config_path)
            mlflow.log_params(params, run_id=run_id)

            mlflow.create_external_model(
                model_type="AI System",
                params=params,
                source_run_id=run_id,
            )
        except Exception as e:
            logger.error(f"Failed to log AI System configuration to run {run_id}: {e}")


logger = logging.getLogger(__name__)



import logging
import mlflow
import os
import semver
from urllib.parse import urljoin

from domino.exceptions import UnsupportedOperationException
from ..authentication import get_auth_by_type
from ._constants import MIN_MLFLOW_VERSION, MIN_DOMINO_VERSION
from ..http_request_manager import _HttpRequestManager

# not thread safe. I am not sure if this will be a problem for users, so am not implementing locking
# it is ok if multiple requests are sent to verify Domino version support
global supported
supported = None


def _get_version_endpoint() -> str:
    return urljoin(os.environ["DOMINO_API_HOST"], "version")


def _get_domino_version() -> str:
    req_manager = _HttpRequestManager(get_auth_by_type())
    version_metadata = req_manager.get(_get_version_endpoint()).json()
    return version_metadata["version"]


def _get_mlflow_version() -> str:
    """
    This makes testing easier
    """
    return mlflow.__version__


def _verify_domino_support_impl():
    domino_supported = True
    domino_version = None
    try:
        # we do our best to get the Domino version. If this code runs in a Domino execution,
        # auth environment variables will be available. If they run this against a local mlflow server/not
        # Domino, it is ok if this check fails
        domino_version = _get_domino_version()
    except Exception as e:
        # the user may run this outside of Domino, so we log a warning instead of failing
        logger.debug(
            f"Failed to get Domino version. Will continue without version info: {e}"
        )

    if domino_version is not None:
        # verify Domino version is >= min domino version
        domino_supported = (
            semver.Version.parse(domino_version).compare(MIN_DOMINO_VERSION) > -1
        )

    if not domino_supported:
        raise UnsupportedOperationException(
            "This version of Domino doesn’t support the aisystems package."
        )

    # verify mlflow sdk version
    mlflow_supported = (
        semver.Version.parse(_get_mlflow_version()).compare(MIN_MLFLOW_VERSION) > -1
    )

    if not mlflow_supported:
        raise UnsupportedOperationException(
            f"This code requires you to install mlflow>={MIN_MLFLOW_VERSION}"
        )


def verify_domino_support():
    global supported
    if supported is None:
        _verify_domino_support_impl()
        supported = True


logger = logging.getLogger(__name__)



"""
Minimum Domino version supported by this python-domino library
"""
MINIMUM_SUPPORTED_DOMINO_VERSION = "4.1.0"

"""
Minimum Domino version allowed by this python-domino library
This version is set to be deprecated in the future
"""
MINIMUM_GA_DOMINO_VERSION = "3.6.0"

"""
Minimum Domino version supporting on demand spark cluster
"""
MINIMUM_ON_DEMAND_SPARK_CLUSTER_SUPPORT_DOMINO_VERSION = "4.2.0"

"""
Minimum Domino version supporting distributed compute cluster launching
"""
MINIMUM_DISTRIBUTED_CLUSTER_SUPPORT_DOMINO_VERSION = "4.5.0"

"""
Distributed compute cluster types and their minimum supported Domino version
"""
CLUSTER_TYPE_MIN_SUPPORT = [
    ("Spark", "4.5.0"),
    ("Ray", "4.5.0"),
    ("Dask", "4.6.0"),
    ("MPI", "5.1.1"),
]

"""
Minimum Domino version that supports compute cluster autoscaling
"""
COMPUTE_CLUSTER_AUTOSCALING_MIN_SUPPORT_DOMINO_VERSION = "5.0.0"

"""
Minimum Domino version supporting external volume mounts
"""
MINIMUM_EXTERNAL_VOLUME_MOUNTS_SUPPORT_DOMINO_VERSION = "4.3.3"

"""
Environment variable names used by this python-domino library. The values
here match the environment variable names a user would find on a deployment.
"""
DOMINO_TOKEN_FILE_KEY_NAME = "DOMINO_TOKEN_FILE"
DOMINO_USER_API_KEY_KEY_NAME = "DOMINO_USER_API_KEY"
DOMINO_HOST_KEY_NAME = "DOMINO_API_HOST"
DOMINO_LOG_LEVEL_KEY_NAME = "DOMINO_LOG_LEVEL"
DOMINO_USER_NAME_KEY_NAME = "DOMINO_USER_NAME"
DOMINO_VERIFY_CERTIFICATE = "DOMINO_VERIFY_CERTIFICATE"
DOMINO_API_PROXY = "DOMINO_API_PROXY"



import logging
import os
import pytest
from unittest.mock import patch

from domino.aisystems._constants import MIN_MLFLOW_VERSION, MIN_DOMINO_VERSION
from domino.aisystems._verify_domino_support import _get_version_endpoint
from domino.exceptions import UnsupportedOperationException
from ..conftest import TEST_AI_SYSTEMS_ENV_VARS

def test_get_version_endpoint():
        with patch.dict(os.environ, TEST_AI_SYSTEMS_ENV_VARS | {"DOMINO_API_HOST": "http://localhost:1111/"}, clear=True):
                assert _get_version_endpoint() == "http://localhost:1111/version"

def test_verify_domino_support_when_get_domino_version_fails(caplog):
        """
        If we fail to get the domino version, we shouldn't fail everything, since this may be due to network error
        and likely if they are on the wrong domino version, the mlflow-proxy won't support new code anyway.
        """
        with patch.dict(os.environ, TEST_AI_SYSTEMS_ENV_VARS | {"DOMINO_API_HOST": "http://localhost:1111/"}, clear=True), \
                patch('domino.aisystems._verify_domino_support._get_domino_version', side_effect=RuntimeError("test_verify_domino_support_when_get_domino_version_fails")), \
                patch('domino.aisystems._verify_domino_support._get_mlflow_version') as mock_get_mlflow_version, \
                caplog.at_level(logging.DEBUG):

                from domino.aisystems._verify_domino_support import _verify_domino_support_impl
                mock_get_mlflow_version.return_value = MIN_MLFLOW_VERSION

                # Should not raise and should pass
                _verify_domino_support_impl()
                assert "Failed to get Domino version. Will continue without version info: test_verify_domino_support_when_get_domino_version_fails" in caplog.text

def test_verify_domino_support_domino_and_mlflow_correct_version(verify_domino_support_fixture):
        from domino.aisystems._verify_domino_support import _verify_domino_support_impl
        verify_domino_support_fixture['mock_get_domino_version'].return_value = MIN_DOMINO_VERSION
        verify_domino_support_fixture['mock_get_mlflow_version'].return_value = MIN_MLFLOW_VERSION

        # Should not raise
        _verify_domino_support_impl()

@pytest.mark.order(1)
def test_verify_domino_support_should_be_idempotent(verify_domino_support_fixture, mocker):
        """
        This test must run first, because if verifies global functionality, which is incidentally exercised
        by other tests.
        """
        from domino.aisystems._verify_domino_support import verify_domino_support
        verify_domino_support_fixture['mock_get_domino_version'].return_value = MIN_DOMINO_VERSION
        verify_domino_support_fixture['mock_get_mlflow_version'].return_value = MIN_MLFLOW_VERSION

        import domino.aisystems._verify_domino_support
        get_domino_version_spy = mocker.spy(domino.aisystems._verify_domino_support, "_get_domino_version")

        verify_domino_support()
        verify_domino_support()

        assert get_domino_version_spy.call_count == 1

def test_verify_domino_support_domino_wrong_version(verify_domino_support_fixture):
        from domino.aisystems._verify_domino_support import _verify_domino_support_impl
        verify_domino_support_fixture['mock_get_domino_version'].return_value = "6.1.2"

        with pytest.raises(UnsupportedOperationException) as exn:
                _verify_domino_support_impl()

        assert str(exn.value) == "This version of Domino doesn’t support the aisystems package."

def test_verify_domino_support_mlflow_wrong_version(verify_domino_support_fixture):
        from domino.aisystems._verify_domino_support import _verify_domino_support_impl
        verify_domino_support_fixture['mock_get_domino_version'].return_value = MIN_DOMINO_VERSION
        verify_domino_support_fixture['mock_get_mlflow_version'].return_value = '3.1.0'

        with pytest.raises(UnsupportedOperationException) as exn:
                _verify_domino_support_impl()

        assert str(exn.value) == f"This code requires you to install mlflow>={MIN_MLFLOW_VERSION}"

@pytest.fixture
def verify_domino_support_fixture():
        with patch.dict(os.environ, TEST_AI_SYSTEMS_ENV_VARS | {"DOMINO_API_HOST": "http://localhost:1111/"}, clear=True), \
                patch('domino.aisystems._verify_domino_support._get_domino_version') as mock_get_domino_version, \
                patch('domino.aisystems._verify_domino_support._get_mlflow_version') as mock_get_mlflow_version:
                yield { 'mock_get_domino_version': mock_get_domino_version, 'mock_get_mlflow_version': mock_get_mlflow_version }



# coding: utf-8

"""
    Domino Public API

    Public API endpoints for Custom Metrics  # noqa: E501

    The version of the OpenAPI document: 5.3.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from domino._impl.custommetrics import schemas  # noqa: F401


class MetricTagV1(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "value",
            "key",
        }
        
        class properties:
            key = schemas.StrSchema
            value = schemas.StrSchema
            __annotations__ = {
                "key": key,
                "value": value,
            }
    
    value: MetaOapg.properties.value
    key: MetaOapg.properties.key
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["key"]) -> MetaOapg.properties.key: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["value"]) -> MetaOapg.properties.value: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["key", "value", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["key"]) -> MetaOapg.properties.key: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["value"]) -> MetaOapg.properties.value: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["key", "value", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *args: typing.Union[dict, frozendict.frozendict, ],
        value: typing.Union[MetaOapg.properties.value, str, ],
        key: typing.Union[MetaOapg.properties.key, str, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'MetricTagV1':
        return super().__new__(
            cls,
            *args,
            value=value,
            key=key,
            _configuration=_configuration,
            **kwargs,
        )



class DominoException(Exception):
    """Base class for Domino Exceptions"""

    pass


class DatasetNotFoundException(DominoException):
    """Dataset not found Exception"""

    pass


class DatasetExistsException(DominoException):
    """Dataset already exists Exception"""

    pass


class RunNotFoundException(DominoException):
    """Run not found Exception"""

    pass


class ProjectNotFoundException(DominoException):
    """Project not found Exception"""

    pass


class RunFailedException(DominoException):
    """Run Failed Exception"""

    pass


class EnvironmentNotFoundException(DominoException):
    """Environment not found Exception"""

    pass


class HardwareTierNotFoundException(DominoException):
    """Hardware tier not found Exception"""

    pass


class CommitNotFoundException(DominoException):
    """Commit not found Exception"""

    pass


class OnDemandSparkClusterNotSupportedException(DominoException):
    """On Demand Spark Cluster not supported"""

    pass


class ExternalVolumeMountsNotSupportedException(DominoException):
    """External Volume Mounts not supported"""

    pass


class UserNotFoundException(DominoException):
    """User not found Exception"""

    pass


class UnsupportedFieldException(DominoException):
    """Unsupported field Exception"""

    pass

class UnsupportedOperationException(DominoException):
    """Unsupported operation Exception"""

    pass

class MalformedInputException(DominoException):
    """Malformed input Exception"""

    pass


class MissingRequiredFieldException(DominoException):
    """Missing required field Exception"""

    pass


class ReloginRequiredException(DominoException):
    """Re-login required Exception"""

    pass



import logging
import os
from typing import Optional
import yaml


def _get_ai_system_config_path() -> str:
    return os.environ.get("DOMINO_AI_SYSTEM_CONFIG_PATH", "./ai_system_config.yaml")


def flatten_dict(d, parent_key="", sep="."):
    """Recursively flattens a nested dictionary."""
    items = []
    for k, v in d.items():
        new_key = f"{parent_key}{sep}{k}" if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def get_flattened_ai_system_config(path: Optional[str] = None) -> dict[str, any]:
    config = read_ai_system_config(path)
    return flatten_dict(config)


def read_ai_system_config(path: Optional[str] = None) -> dict:
    """For getting the ai_system_config.yaml file and reading it into a dictionary.
    If no path is provided it will look for the path in the DOMINO_AI_SYSTEM_CONFIG_PATH. See environment variables
    docs for default values.

    Args:
        path: Location of the ai system config yaml file.

    Returns:
        A dictionary of the ai system config yaml values.

    """
    path = path or _get_ai_system_config_path()
    params = {}
    try:
            with open(path, 'r') as f:
                params = yaml.safe_load(f)
    except Exception as e:
            logging.warning(f"Failed to read ai system config yaml at path {path}: {e}")

    return params


logger = logging.getLogger(__name__)



import re
from typing import Optional

from ._constants import EVALUATION_TAG_PREFIX
from ..exceptions import DominoException

VALID_LABEL_PATTERN = r"[a-zA-Z0-9_-]+"
TAG_MATCHER_PATTERN = (
    rf"^{EVALUATION_TAG_PREFIX}\.(metric|label)\.({VALID_LABEL_PATTERN})$"
)


class InvalidEvaluationLabelException(DominoException):
    """Invalid evaluation label Exception"""

    pass


def validate_label(label: str):
    if not re.match(VALID_LABEL_PATTERN, label):
        raise InvalidEvaluationLabelException(
            f"label '{label}' may contain only alphanumeric characters, underscores and dashes."
        )


def get_eval_tag_name(mlflow_tag: str) -> Optional[str]:
    match = re.match(TAG_MATCHER_PATTERN, mlflow_tag)
    if match:
        return match.group(2)
    return None


def build_metric_tag(metric_name: str) -> str:
    return f"{EVALUATION_TAG_PREFIX}.metric.{metric_name}"


def build_eval_result_tag(label: str, result) -> str:
    try:
        float(result)
        return build_metric_tag(label)
    except ValueError:
        # this result will be treated as a string
        # build label tag
        return f"{EVALUATION_TAG_PREFIX}.label.{label}"



import logging
import os
from http import HTTPStatus

import requests
from bs4 import BeautifulSoup
from requests.adapters import HTTPAdapter, Retry
from requests.auth import AuthBase

from ._version import __version__
from .constants import DOMINO_VERIFY_CERTIFICATE
from .exceptions import ReloginRequiredException


R_SESSION_MAX_RETRIES = 4

class _SessionInitializer:
    def __initialize__(self, session):
        raise NotImplementedError('Session initializers must be callable.')

class _HttpRequestManager:
    """
    This class is responsible for
    making Http request calls
    """

    def __init__(self, auth: AuthBase):
        self.auth = auth
        self._logger = logging.getLogger(__name__)
        self.request_session = self._set_session()

        if isinstance(self.auth, _SessionInitializer):
            self.auth.__initialize__(self._set_session())

    def _set_session(self):
        """
        Initialize a request session with retry to help manage connections drop.
        """
        self.session = requests.Session()
        retries = Retry(
            total=int(os.getenv("DOMINO_MAX_RETRIES", str(R_SESSION_MAX_RETRIES))),
            backoff_factor=1,  # retry seconds
            status_forcelist=[408, 502, 503, 504],
        )

        if os.environ.get(DOMINO_VERIFY_CERTIFICATE, None) in ["false", "f", "n", "no"]:
            warning = "InsecureRequestWarning: Bypassing certificate verification is strongly ill-advised"
            logging.warning(warning)
            print(warning)
            self.session.verify = False

        retry_adapter = HTTPAdapter(max_retries=retries)
        self.session.mount("https://", retry_adapter)
        return self.session

    def post(self, url, data=None, json=None, **kwargs):
        return self._raise_for_status(
            self.request_session.post(
                url, auth=self.auth, data=data, json=json, **kwargs
            )
        )

    def patch(self, url, data=None, json=None, **kwargs):
        return self._raise_for_status(
            self.request_session.patch(
                url, auth=self.auth, data=data, json=json, **kwargs
            )
        )

    def get(self, url, **kwargs):
        return self._raise_for_status(
            self.request_session.get(url, auth=self.auth, **kwargs)
        )

    def put(self, url, data=None, **kwargs):
        return self._raise_for_status(
            self.request_session.put(url, auth=self.auth, data=data, **kwargs)
        )

    def delete(self, url, **kwargs):
        return self._raise_for_status(
            self.request_session.delete(url, auth=self.auth, **kwargs)
        )

    def get_raw(self, url):
        return self.get(url, stream=True).raw

    def _raise_for_status(self, response):
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == HTTPStatus.CONFLICT:
                raise ReloginRequiredException("Temporary credentials expired")
            # Sometimes, the error response is a long HTML page.
            # We don't want to log error the whole response html in those cases.
            if not bool(BeautifulSoup(e.response.text, "html.parser").find()):
                self._logger.error(e.response.text)
            else:
                self._logger.debug(e.response.text)
            raise
        return response



import pathlib
import re
import sys

from setuptools import find_packages, setup

if sys.version_info < (3, 10):
    message = f"dominodatalab requires Python '>=3.10.0' but the running Python is {'.'.join(map(str,sys.version_info[:3]))}"
    message += "\nConsider Checking python-domino and domino compatibility"
    sys.exit(message)

PACKAGE_NAME = "domino"

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()


def get_version():
    try:
        f = open(f"{PACKAGE_NAME}/_version.py")
    except EnvironmentError:
        return None
    for line in f.readlines():
        mo = re.match('__version__ = "([^\']+)"', line)
        if mo:
            ver = mo.group(1)
            return ver
    return None


setup(
    name="dominodatalab",
    version=get_version(),
    author="Domino Data Lab",
    author_email="support@dominodatalab.com",
    packages=find_packages(),
    scripts=[],
    url="https://github.com/dominodatalab/python-domino",
    download_url=f"https://github.com/dominodatalab/python-domino/archive/Release-{get_version()}.zip",
    license="Apache Software License (Apache 2.0)",
    description="Python bindings for the Domino API",
    long_description=README,
    long_description_content_type="text/markdown",
    keywords=["Domino Data Lab", "API"],
    python_requires=">=3.10.0",
    install_requires=[
        "packaging==23.2",
        "requests>=2.4.2",
        "beautifulsoup4~=4.11",
        "polling2~=0.5.0",
        "urllib3>=1.26.19,<3",
        "typing-extensions~=4.13.0",
        "frozendict~=2.3",
        "python-dateutil~=2.8.2",
        "retry==0.9.2",
    ],
    extras_require={
        "airflow": ["apache-airflow==2.2.4"],
        "data": ["dominodatalab-data>=0.1.0"],
        "aisystems": [
            "semver>=3.0.4",
            "pandas>=2.3.1",
            "numpy>=2.0.2",
            "mlflow-skinny~=3.2",
            "mlflow-tracing~=3.2",
        ],
        "dev": [
            "pytest-order>=1.3.0",
            "pytest-asyncio>=0.23.8",
            "scikit-learn>=1.6.1",  # used in aisystem tracing tests
            "black==22.3.0",
            "flake8==4.0.1",
            "Jinja2==2.11.3",
            "nbconvert==6.3.0",
            "packaging==23.2",
            "polling2==0.5.0",
            "pre-commit==2.19.0",
            "pyspark==3.3.0",
            "pytest==7.4.3",
            "requests_mock==1.9.3",
            "tox==3.25.1",
            "frozendict==2.4.6",
            "docker>=7.1.0",
            "pytest-mock>=3.14.1",
        ],
        "docs": [
            "sphinx>=7.4.0",
            "markupsafe==2.0.1", # added for using Jinja2 with sphinx and python 3.10
        ]
    },
)



# this file is for documenting environment variables and has no functional value

"""
:DOMINO_AI_SYSTEM_CONFIG_PATH: For configuring the location of the ai_system_config.yaml file.
    If not set, defaults to './ai_system_config.yaml'.
:type: str

:DOMINO_AI_SYSTEM_MODEL_ID: The ID of the production AI System
:type: str
"""



from unittest.mock import patch
import os

from domino.aisystems import read_ai_system_config
from domino.aisystems.read_ai_system_config import flatten_dict
from ..conftest import TEST_AI_SYSTEMS_ENV_VARS

def test_read_ai_system_config_path_from_env_var():
        with patch.dict(os.environ, TEST_AI_SYSTEMS_ENV_VARS, clear=True):
                config_values = read_ai_system_config()

                assert config_values['version'] == 1.0
                assert config_values['chat_assistant']['model'] == 'gpt-3.5-turbo'
                assert config_values['chat_assistant']['temperature'] == 0.7
                assert config_values['chat_assistant']['max_tokens'] == 1500

def test_read_ai_system_config_path_from_override_arg():
        with patch.dict(os.environ, TEST_AI_SYSTEMS_ENV_VARS | {"DOMINO_AI_SYSTEM_CONFIG_PATH": "broken_path"}, clear=True):
                config_values = read_ai_system_config("tests/assets/ai_system_config.yaml")

                assert config_values['version'] == 1.0
                assert config_values['chat_assistant']['model'] == 'gpt-3.5-turbo'
                assert config_values['chat_assistant']['temperature'] == 0.7
                assert config_values['chat_assistant']['max_tokens'] == 1500

def test_flatten_dict():
        nested_dict = {
                'a': 1,
                'b': {
                        'c': 2,
                        'd': { 'e': 3 }
                },
                'f': 4
        }
        flat_dict = flatten_dict(nested_dict)
        expected_flat_dict = {
                'a': 1,
                'b.c': 2,
                'b.d.e': 3,
                'f': 4
        }
        assert flat_dict == expected_flat_dict



# coding: utf-8

"""
    Domino Public API

    Public API endpoints for Custom Metrics  # noqa: E501

    The version of the OpenAPI document: 5.3.0
    Generated by: https://openapi-generator.tech
"""

from domino._impl.custommetrics.paths.api_metric_values_v1.post import LogMetricValues
from domino._impl.custommetrics.paths.api_metric_values_v1_model_monitoring_id_metric.get import RetrieveMetricValues
from domino._impl.custommetrics.paths.api_metric_alerts_v1.post import SendMetricAlert


class CustomMetricsApi(
    LogMetricValues,
    RetrieveMetricValues,
    SendMetricAlert,
):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    pass



import os
import re

from requests.auth import AuthBase, HTTPBasicAuth

from .constants import DOMINO_TOKEN_FILE_KEY_NAME, DOMINO_USER_API_KEY_KEY_NAME, \
    DOMINO_API_PROXY


class ProxyAuth(AuthBase):
    """
    Class for authenticating requests using the Domino API reverse Proxy.
    All Domino URLs will use the proxy as host instead.
    """

    def __init__(self, api_proxy):
        match = re.search("(https?://)?([^/]+:[0-9]+)$", api_proxy)
        if not match:
            raise RuntimeError("Bad proxy URL: '%s', must be host:port or scheme://host:port" % api_proxy)
        if not match.group(1):
            proxy_str = "http://" + match.group(2)
        else:
            proxy_str = api_proxy
        self.api_proxy = proxy_str

    def __call__(self, r):
        """
        Override the default __call__ method for the AuthBase base class.
        Note this override is no-op, the actual work happens in the initialize method.

        More more info, see:
        https://docs.python-requests.org/en/master/user/advanced/
        """
        r.url = self._replaceHostWithProxy(r.url)
        return r

    def _replaceHostWithProxy(self, url):
        return re.sub('^.*?://[^/]+', self.api_proxy, url)



class BearerAuth(AuthBase):
    """
    Class for authenticating requests by user supplied token.
    """

    def __init__(self, auth_token=None, domino_token_file=None):
        self.auth_token = auth_token
        self.domino_token_file = domino_token_file

    def _from_token_file(self):
        with open(self.domino_token_file, "r") as token_file:
            return token_file.readline().rstrip()

    def __call__(self, r):
        """
        Override the default __call__ method for the AuthBase base class

        More more info, see:
        https://docs.python-requests.org/en/master/user/advanced/
        """
        auth_token = (
            self._from_token_file() if self.domino_token_file else self.auth_token
        )
        r.headers["Authorization"] = "Bearer " + auth_token
        return r


def get_auth_by_type(api_key=None, auth_token=None, domino_token_file=None, api_proxy=None):
    """
    Return appropriate authentication object for requests.

    If no authentication credential is provided, the call fails with an AssertError

    Precedence in the case of multiple credentials is:
        1. api_proxy
        2. auth_token string
        3. domino_token_file
        4. api_key
        5. api_proxy_from_env
        6. domino_token_file_from_env
        7. api_key_from_env
    """

    if api_proxy:
        return ProxyAuth(api_proxy)
    elif auth_token:
        return BearerAuth(auth_token=auth_token)
    elif domino_token_file:
        return BearerAuth(domino_token_file=domino_token_file)
    elif api_key:
        return HTTPBasicAuth("", api_key)
    else:
        # In the case that no authentication type was passed when this method
        # called, fall back to deriving the auth info from the environment.
        api_proxy_from_env = os.getenv(DOMINO_API_PROXY)
        api_key_from_env = os.getenv(DOMINO_USER_API_KEY_KEY_NAME)
        domino_token_file_from_env = os.getenv(DOMINO_TOKEN_FILE_KEY_NAME)
        if api_key_from_env or domino_token_file_from_env or api_proxy_from_env:
            return get_auth_by_type(
                api_key=api_key_from_env, domino_token_file=domino_token_file_from_env, api_proxy=api_proxy_from_env
            )
        else:
            # All attempts failed -- nothing to do but raise an error.
            raise RuntimeError("Unable to authenticate: no authentication provided")



from domino.aisystems._eval_tags import build_eval_result_tag

def test_build_eval_result_tags():
        assert build_eval_result_tag('my_metric', '1') ==  'domino.prog.metric.my_metric', 'numbers should be metrics'
        assert build_eval_result_tag('my_label', 'cat') ==  'domino.prog.label.my_label', 'strings should be labels'



# flake8: noqa

_import_error_message = (
    "domino.data_sources is not installed.\n\n"
    "Please pip install dominodatalab-data:\n\n"
    "  python -m pip install dominodatalab-data   # install directly\n"
    '  python -m pip install "dominodatalab[data]" --upgrade   # install via extra dependency'
)

try:
    from domino_data.data_sources import *
except ImportError as e:
    if e.msg == "No module named 'domino_data'":
        raise ImportError(_import_error_message) from e
    else:
        raise


def __getattr__(value):
    try:
        import domino_data.data_sources
    except ImportError as e:
        raise ImportError(_import_error_message) from e
    return getattr(domino_data.data_sources, value)



import os
import socket
import urllib

from packaging.version import parse as parse_version

from . import constants


def is_version_compatible(version: str) -> bool:
    """
    Helper function to check for version compatibility

    @:param version  Domino version to check version compatibility against
    @:return bool   Boolean representing if version is compatible or not
    """
    return parse_version(version) >= parse_version(
        constants.MINIMUM_SUPPORTED_DOMINO_VERSION
    )


def is_compute_cluster_autoscaling_supported(version: str) -> bool:
    return parse_version(version) >= parse_version(
        constants.COMPUTE_CLUSTER_AUTOSCALING_MIN_SUPPORT_DOMINO_VERSION
    )


def is_cluster_type_supported(version: str, cluster_type: str) -> bool:
    curr_version = parse_version(version)

    return next(
        (
            True
            for ct, min_version in constants.CLUSTER_TYPE_MIN_SUPPORT
            if ct == cluster_type and curr_version >= parse_version(min_version)
        ),
        False,
    )


def is_compute_cluster_properties_supported(version: str) -> bool:
    return parse_version(version) >= parse_version(
        constants.MINIMUM_DISTRIBUTED_CLUSTER_SUPPORT_DOMINO_VERSION
    )


def is_on_demand_spark_cluster_supported(version: str) -> bool:
    return parse_version(version) >= parse_version(
        constants.MINIMUM_ON_DEMAND_SPARK_CLUSTER_SUPPORT_DOMINO_VERSION
    )


def is_external_volume_mounts_supported(version: str) -> bool:
    return parse_version(version) >= parse_version(
        constants.MINIMUM_EXTERNAL_VOLUME_MOUNTS_SUPPORT_DOMINO_VERSION
    )


def clean_host_url(host_url):
    """
    Helper function to clean 'host_url'. This will extract
    hostname (with scheme) from the url
    """
    url_split = urllib.parse.urlsplit(host_url)
    return f"{url_split.scheme}://{url_split.netloc}"


def domino_is_reachable(url=os.getenv(constants.DOMINO_HOST_KEY_NAME), port="443"):
    """
    Confirm that a deployment is accessible.

    Returns Boolean value.
    """
    if url is None:
        return False

    fqdn = urllib.parse.urlsplit(url).netloc
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((fqdn, int(port)))
        is_reachable = True
    except OSError:
        print(f"{fqdn}:{port} is not reachable")
        is_reachable = False
    finally:
        s.close()

    return is_reachable



"""
To run these tests you need to have at least configured airflow in
local mode and run:

`airflow db init`
"""
import os
from datetime import datetime

import pytest
from airflow.operators.dummy import DummyOperator

from domino.airflow import DominoOperator
from domino.exceptions import RunFailedException
from airflow import DAG
from airflow.models import TaskInstance


TEST_PROJECT = os.environ.get("DOMINO_TEST_PROJECT")
dag_id = "test_operator"


def test_airflow_dags():
    pytest.importorskip("airflow")

    from airflow import DAG
    from airflow.models import TaskInstance

    start_time = datetime.now()

    dag = DAG(dag_id, start_date=start_time)
    task = DummyOperator(
        dag=dag,
        task_id='test_airflow_dags',
    )

    task.run()
    ti = TaskInstance(task=task, execution_date=start_time)
    task.execute(ti.get_template_context())


def test_operator():
    start_time = datetime.now()

    dag = DAG(dag_id, start_date=start_time)
    task = DominoOperator(
        dag=dag,
        task_id="test_operator",
        project=TEST_PROJECT,
        isDirect=True,
        command=["python -V"],
    )

    task.run()
    ti = TaskInstance(task=task, execution_date=start_time)
    task.execute(ti.get_template_context())


def test_operator_fail(caplog):
    execution_dt = datetime.now()

    dag = DAG(dag_id, start_date=execution_dt)
    task = DominoOperator(
        dag=dag,
        task_id="test_operator_fail",
        project=TEST_PROJECT,
        isDirect=True,
        command=["python -c 'import sys; sys.exit(1)'"],
    )

    with pytest.raises(RunFailedException):
        task.run()
        ti = TaskInstance(task=task, execution_date=execution_dt)
        task.execute(ti.get_template_context())


def test_operator_fail_invalid_tier(caplog):
    execution_dt = datetime.now()

    dag = DAG(dag_id, start_date=execution_dt)
    task = DominoOperator(
        dag=dag,
        task_id="test_operator_fail_invalid_tier",
        project=TEST_PROJECT,
        isDirect=True,
        command=["python -V"],
        tier="this tier does not exist",
    )

    with pytest.raises(ValueError):
        task.run()
        ti = TaskInstance(task=task, execution_date=execution_dt)
        task.execute(ti.get_template_context())



import functools
import json
import logging
import os
from packaging import version
import re
import time
from typing import Any, Dict, List, Optional, Tuple
import warnings

import polling2
import requests
from bs4 import BeautifulSoup

from domino import exceptions, helpers, datasets
from domino._version import __version__
from domino.authentication import get_auth_by_type
from domino.domino_enums import BillingTagSettingMode, BudgetLabel, BudgetType, ProjectVisibility
from domino.constants import (
    CLUSTER_TYPE_MIN_SUPPORT,
    DOMINO_HOST_KEY_NAME,
    DOMINO_LOG_LEVEL_KEY_NAME,
    MINIMUM_EXTERNAL_VOLUME_MOUNTS_SUPPORT_DOMINO_VERSION,
    MINIMUM_ON_DEMAND_SPARK_CLUSTER_SUPPORT_DOMINO_VERSION,
    MINIMUM_SUPPORTED_DOMINO_VERSION,
)
from domino.http_request_manager import _HttpRequestManager
from domino.routes import _Routes
from domino._custom_metrics import _CustomMetricsClientBase, _CustomMetricsClientGen, _CustomMetricsClient


class Domino:
    def __init__(
        self, project, api_key=None, host=None, domino_token_file=None, auth_token=None, api_proxy=None,
    ):

        self._configure_logging()

        _host = host or os.getenv(DOMINO_HOST_KEY_NAME)
        assert _host, (
            "Host must be supplied as a parameter or through the "
            f"{DOMINO_HOST_KEY_NAME} environment variable."
        )
        host = helpers.clean_host_url(_host)

        try:
            self._owner_username, self._project_name = project.split("/")
            self._routes = _Routes(host, self._owner_username, self._project_name)
        except ValueError:
            self._logger.error(
                f"Project {project} must be given in the form username/projectname"
            )
            raise

        # This call sets self.request_manager
        self.authenticate(api_key, auth_token, domino_token_file, api_proxy)

        # Get version
        self._version = self.deployment_version().get("version")
        assert self.requires_at_least(MINIMUM_SUPPORTED_DOMINO_VERSION)

        self._logger.debug(
            f"Domino deployment {host} is running version {self._version}"
        )

        # Check version compatibility
        if not helpers.is_version_compatible(self._version):
            error_message = (
                f"Domino version: {self._version} is not compatible with "
                f"python-domino version: {__version__}"
            )
            self._logger.error(error_message)
            raise Exception(error_message)

    @property
    def log(self):
        try:
            return self._logger
        except AttributeError:
            self._configure_logging()
            return self._logger

    def _configure_logging(self):
        logging_level = logging.getLevelName(
            os.getenv(DOMINO_LOG_LEVEL_KEY_NAME, "INFO").upper()
        )
        logging.basicConfig(level=logging_level)
        self._logger = logging.getLogger(__name__)

    def authenticate(self, api_key=None, auth_token=None, domino_token_file=None, api_proxy=None):
        """
        Method to authenticate the request manager. An existing domino client object can
        use this with a new token if the existing credentials expire.
        """
        self.request_manager = _HttpRequestManager(
            get_auth_by_type(
                api_key=api_key,
                auth_token=auth_token,
                domino_token_file=domino_token_file,
                api_proxy=api_proxy,
            )
        )

    def commits_list(self):
        url = self._routes.commits_list()
        return self._get(url)

    def runs_list(self):
        url = self._routes.runs_list()
        return self._get(url)

    def runs_start(
        self,
        command,
        isDirect=False,
        commitId=None,
        title=None,
        tier=None,
        publishApiEndpoint=None,
    ):

        url = self._routes.runs_start()

        request = {
            "command": command,
            "isDirect": isDirect,
            "commitId": commitId,
            "title": title,
            "tier": tier,
            "publishApiEndpoint": publishApiEndpoint,
        }
        try:
            response = self.request_manager.post(url, json=request)
            return response.json()
        except exceptions.ReloginRequiredException:
            self._logger.info(
                f" You need to log in to the Domino UI to start the run. Please do it at {self._routes.host}/relogin?redirectPath=/"
            )

    def runs_start_blocking(
        self,
        command,
        isDirect=False,
        commitId=None,
        title=None,
        tier=None,
        publishApiEndpoint=None,
        poll_freq=5,
        max_poll_time=6000,
        retry_count=5,
    ):
        """
        Run a tasks that runs in a blocking loop that periodically checks to
        see if the task is done.  If the task errors an exception is raised.

        parameters
        ----------
        command : list of strings
                  list that containst the name of the file to run in index 0
                  and args in subsequent positions.
                  example:
                  >> domino.runs_start(["main.py", "arg1", "arg2"])

        isDirect : boolean (Optional)
                   Whether or not this command should be passed directly to
                   a shell.

        commitId : string (Optional)
                   The commitId to launch from. If not provided, will launch
                   from latest commit.

        title    : string (Optional)
                   A title for the run

        tier     : string (Optional)
                   The hardware tier to use for the run. Will use project
                   default tier if not provided.

        publishApiEndpoint : boolean (Optional)
                            Whether or not to publish an API endpoint from the
                            resulting output.

        poll_freq : int (Optional)
                    Number of seconds in between polling of the Domino server
                    for status of the task that is running.

        max_poll_time : int (Optional)
                        Maximum number of seconds to wait for a task to
                        complete. If this threshold is exceeded, an exception
                        is raised.

        retry_count : int (Optional)
                        Maximum number of retry to do while polling
                        (in-case of transient http errors). If this
                        threshold exceeds, an exception is raised.
        """
        run_response = self.runs_start(
            command, isDirect, commitId, title, tier, publishApiEndpoint
        )
        run_id = run_response["runId"]

        poll_start = time.time()
        current_retry_count = 0
        while True:
            try:
                run_info = self.get_run_info(run_id)
                if run_info is None:
                    raise exceptions.RunNotFoundException(
                        f"Tried to access nonexistent run id {run_id}"
                    )
                current_retry_count = 0
            except (
                requests.exceptions.RequestException,
                exceptions.RunNotFoundException,
            ) as e:
                current_retry_count += 1
                self.log.warning(f"Failed to get run info for runId: {run_id} : {e}")
                if current_retry_count > retry_count:
                    raise Exception(
                        f"Cannot get run info, max retry {retry_count} exceeded"
                    ) from None
                else:
                    self.log.info(
                        f"Retrying ({current_retry_count}/{retry_count}) getting run info ..."
                    )
                    time.sleep(poll_freq)
                    continue

            elapsed_time = time.time() - poll_start

            if elapsed_time >= max_poll_time:
                raise Exception(f"Run exceeded maximum time of {max_poll_time} seconds")

            output_commit_id = run_info.get("outputCommitId")
            if not output_commit_id:
                time.sleep(poll_freq)
                continue

            # once task has finished running check to see if it was successful
            else:
                stdout_msg = self.get_run_log(runId=run_id, includeSetupLog=False)

                if run_info["status"] != "Succeeded":
                    self.process_log(stdout_msg)
                    raise exceptions.RunFailedException(
                        f"Remote run {run_id} finished but did not succeed."
                    )

                break

        return run_response

    def run_stop(self, runId, saveChanges=True):
        self.log.warning("Use job_stop method instead")
        return self.job_stop(job_id=runId, commit_results=saveChanges)

    def runs_status(self, runId):
        url = self._routes.runs_status(runId)
        return self._get(url)

    def get_run_log(self, runId, includeSetupLog=True):
        """
        Get the unified log for a run (setup + stdout).

        parameters
        ----------
        runId : string
                the id associated with the run.
        includeSetupLog : bool
                whether or not to include the setup log in the output.
        """

        url = self._routes.runs_stdout(runId)

        logs = list()

        if includeSetupLog:
            logs.append(self._get(url)["setup"])

        logs.append(self._get(url)["stdout"])

        return "\n".join(logs)

    def get_run_info(self, run_id):
        for run_info in self.runs_list()["data"]:
            if run_info["id"] == run_id:
                return run_info

    def runs_stdout(self, runId):
        """
        Get std out emitted by a particular run.

        parameters
        ----------
        runId : string
                the id associated with the run.
        """

        html_start_tags = (
            "<pre style='white-space: pre-wrap; white-space: -moz-pre-wrap; white-space: -pre-wrap; "
            "white-space: -o-pre-wrap; word-wrap: break-word; word-wrap: break-all;'>"
        )
        html_end_tags = "</pre>$"
        span_regex = re.compile("<?span.*?>")
        returns = "'\\n'\n"

        url = self._routes.runs_stdout(runId)
        raw_stdout = self._get(url)["stdout"]

        stdout = (
            re.sub(html_end_tags, "", re.sub(span_regex, "", raw_stdout))
            .replace(html_start_tags, "")
            .replace(returns, "\n")
        )

        return stdout

    def job_start(
        self,
        command: str,
        commit_id: Optional[str] = None,
        hardware_tier_id: Optional[str] = None,
        hardware_tier_name: Optional[str] = None,
        environment_id: Optional[str] = None,
        on_demand_spark_cluster_properties: Optional[dict] = None,
        compute_cluster_properties: Optional[dict] = None,
        external_volume_mounts: Optional[List[str]] = None,
        title: Optional[str] = None,
    ) -> dict:
        """
        Starts a Domino Job via V4 API
        :param command:                             string
                                                    Command to execute in Job
                                                    >> domino.job_start(command="main.py arg1 arg2")
        :param commit_id:                           string (Optional)
                                                    The commitId to launch from. If not provided, will launch
                                                    from latest commit.
        :param hardware_tier_id:                    string (Optional)
                                                    The hardware tier ID to launch job in. If not provided
                                                    it will use the default hardware tier for the project
        :param hardware_tier_name:                  string (Optional)
                                                    This field is deprecated. Please use hardware_tier_id.
                                                    The hardware tier NAME to launch job in. If not provided
                                                    it will use the default hardware tier for the project
        :param environment_id:                      string (Optional)
                                                    The environment id to launch job with. If not provided
                                                    it will use the default environment for the project
        :param on_demand_spark_cluster_properties:  dict (Optional)
                                                    This field is deprecated. Please use compute_cluster_properties.
                                                    The following properties can be provided in spark cluster
                                                    {
                                                        "computeEnvironmentId": "<Environment ID configured with spark>"
                                                        "executorCount": "<Number of Executors in cluster>"
                                                         (optional defaults to 1)
                                                        "executorHardwareTierId": "<Hardware tier ID for Spark Executors>"
                                                         (optional defaults to last used historically if available)
                                                        "masterHardwareTierId":  "<Hardware tier ID for Spark master"
                                                         (optional defaults to last used historically if available)
                                                        "executorStorageMB": "<Executor's storage in MB>"
                                                         (optional defaults to 0; 1GB is 1000MB Here)
                                                    }
        :param compute_cluster_properties:          dict (Optional)
                                                    The compute cluster properties definition contains parameters for
                                                    launching any Domino supported compute cluster for a job. Use this
                                                    to launch a job that uses compute cluster instead of
                                                    the deprecated on_demand_spark_cluster_properties field. If
                                                    on_demand_spark_cluster_properties and compute_cluster_properties
                                                    are both present, on_demand_spark_cluster_properties will be ignored.

                                                    compute_cluster_properties contains the following fields:
                                                    {
                                                        "clusterType": <string, one of "Ray", "Spark", "Dask", "MPI">,
                                                        "computeEnvironmentId": <string, The environment ID for the cluster's nodes>,
                                                        "computeEnvironmentRevisionSpec": <one of "ActiveRevision", "LatestRevision",
                                                        {"revisionId":"<environment_revision_id>"} (optional)>,
                                                        "masterHardwareTierId": <string, the Hardware tier ID for the cluster's master node
                                                        (required unless clusterType is MPI)>,
                                                        "workerCount": <number, the total workers to spawn for the cluster>,
                                                        "workerHardwareTierId": <string, The Hardware tier ID for the cluster workers>,
                                                        "workerStorage": <{ "value": <number>, "unit": <one of "GiB", "MB"> },
                                                        The disk storage size for the cluster's worker nodes (optional)>
                                                        "maxWorkerCount": <number, The max number of workers allowed. When
                                                        this configuration exists, autoscaling is enabled for the cluster and
                                                        'workerCount' is interpreted as the min number of workers allowed in the cluster
                                                        (optional)>
                                                    }
        :param external_volume_mounts:              list of string (Optional)
                                                    External volume mount ids to mount to run. If not provided will launch with
                                                    no external volume mounts mounted.
        :param title                                string (Optional)
                                                    Title for the Job
        :return: Returns created Job details (number, id etc)
        """

        def validate_on_demand_spark_cluster_properties(max_execution_slot_per_user):
            self.log.debug(
                f"Validating spark cluster properties: {on_demand_spark_cluster_properties}"
            )
            validations = {
                "executorCount": lambda ec: validate_spark_executor_count(
                    int(ec), max_execution_slot_per_user
                ),
                "computeEnvironmentId": lambda com_env_id: self._validate_environment_id(
                    com_env_id
                ),
                "executorHardwareTierId": lambda exec_hwd_tier_id: self._validate_hardware_tier_id(
                    exec_hwd_tier_id
                ),
                "masterHardwareTierId": lambda master_hwd_tier_id: self._validate_hardware_tier_id(
                    master_hwd_tier_id
                ),
                "executorStorageMB": lambda exec_storage_mb: True,
            }
            if "computeEnvironmentId" not in on_demand_spark_cluster_properties:
                raise exceptions.MissingRequiredFieldException(
                    "Mandatory field computeEnvironmentId not passed in spark properties"
                )
            for k, v in on_demand_spark_cluster_properties.items():
                if not validations.get(k, lambda x: False)(v):
                    raise exceptions.MalformedInputException(
                        f"Invalid spark property {k}:{v} found"
                    )

        def validate_spark_executor_count(executor_count, max_executor_count):
            if executor_count <= max_executor_count:
                return True
            else:
                raise exceptions.MalformedInputException(
                    f"executor count: {executor_count} exceeding max executor count: {max_executor_count}"
                )

        def get_default_spark_settings():
            self.log.debug("Getting default spark settings")
            default_spark_setting_url = self._routes.default_spark_setting(
                self.project_id
            )
            return self.request_manager.get(default_spark_setting_url).json()

        def validate_is_on_demand_spark_supported():
            if not helpers.is_on_demand_spark_cluster_supported(self._version):
                raise exceptions.OnDemandSparkClusterNotSupportedException(
                    f"Your domino deployment version {self._version} does not support on demand spark cluster. "
                    f"Minimum support version {MINIMUM_ON_DEMAND_SPARK_CLUSTER_SUPPORT_DOMINO_VERSION}"
                )

        def validate_distributed_compute_cluster_properties():
            if not helpers.is_compute_cluster_properties_supported(self._version):
                raise exceptions.UnsupportedFieldException(
                    f"'compute_cluster_properties' is not supported in Domino {self._version}."
                )

            required_keys = [
                "clusterType",
                "computeEnvironmentId",
                "masterHardwareTierId",
                "workerHardwareTierId",
                "workerCount",
            ]
            required_key_overrides = {("masterHardwareTierId", "MPI"): False}

            for key in required_keys:
                key_required = required_key_overrides.get(
                    (key, compute_cluster_properties["clusterType"]), True
                )

                if key_required and (key not in compute_cluster_properties):
                    raise exceptions.MissingRequiredFieldException(
                        f"{key} is required in compute_cluster_properties"
                    )

            if not helpers.is_cluster_type_supported(
                self._version, compute_cluster_properties["clusterType"]
            ):
                supported_types = [
                    ct
                    for ct, min_version in CLUSTER_TYPE_MIN_SUPPORT
                    if helpers.is_cluster_type_supported(self._version, ct)
                ]
                supported_types_str = ", ".join(supported_types)
                raise exceptions.MalformedInputException(
                    f"Domino {self._version} does not support cluster type {compute_cluster_properties['clusterType']}."
                    + f" This version of Domino supports the following cluster types: {supported_types_str}"
                )

            def throw_if_information_invalid(key: str, info: dict) -> bool:
                try:
                    self._validate_information_data_type(info)
                except Exception as e:
                    raise exceptions.MalformedInputException(
                        f"{key} in compute_cluster_properties failed validation: {e}"
                    )

            if "workerStorage" in compute_cluster_properties:
                throw_if_information_invalid(
                    "workerStorage", compute_cluster_properties["workerStorage"]
                )

            if compute_cluster_properties["workerCount"] < 1:
                raise exceptions.MalformedInputException(
                    "compute_cluster_properties.workerCount must be greater than 0"
                )

            if (
                "maxWorkerCount" in compute_cluster_properties
                and not helpers.is_compute_cluster_autoscaling_supported(self._version)
            ):
                raise exceptions.UnsupportedFieldException(
                    f"'maxWorkerCount' is not supported in Domino {self._version}."
                )

            if "masterHardwareTierId" in compute_cluster_properties:
                self._validate_hardware_tier_id(compute_cluster_properties["masterHardwareTierId"])

            self._validate_hardware_tier_id(compute_cluster_properties["workerHardwareTierId"])

        def validate_is_external_volume_mounts_supported():
            if not helpers.is_external_volume_mounts_supported(self._version):
                raise exceptions.ExternalVolumeMountsNotSupportedException(
                    f"Your domino deployment version {self._version} does not support external volume mounts. "
                    f"Minimum support version {MINIMUM_EXTERNAL_VOLUME_MOUNTS_SUPPORT_DOMINO_VERSION}"
                )

        spark_cluster_properties = None
        validated_compute_cluster_properties = None

        if commit_id is not None:
            self._validate_commit_id(commit_id)
        if hardware_tier_id is not None:
            self._validate_hardware_tier_id(hardware_tier_id)
        if hardware_tier_name is not None:
            self._validate_hardware_tier_name(hardware_tier_name)
        if environment_id is not None:
            self._validate_environment_id(environment_id)
        if external_volume_mounts is not None:
            validate_is_external_volume_mounts_supported()
        if compute_cluster_properties is not None:
            validate_distributed_compute_cluster_properties()

            validated_compute_cluster_properties = compute_cluster_properties.copy()

            if "masterHardwareTierId" in compute_cluster_properties:
                validated_compute_cluster_properties["masterHardwareTierId"] = {
                    "value": compute_cluster_properties["masterHardwareTierId"]
                }

            validated_compute_cluster_properties["workerHardwareTierId"] = {
                "value": compute_cluster_properties["workerHardwareTierId"]
            }

        elif on_demand_spark_cluster_properties is not None:
            validate_is_on_demand_spark_supported()
            default_spark_setting = get_default_spark_settings()
            max_execution_slot = default_spark_setting["maximumExecutionSlotsPerUser"]
            default_executor_hardware_tier = default_spark_setting[
                "executorHardwareTierId"
            ]
            default_master_hardware_tier = default_spark_setting["masterHardwareTierId"]
            validate_on_demand_spark_cluster_properties(max_execution_slot)
            executor_hardware_tier_id = on_demand_spark_cluster_properties.get(
                "executorHardwareTierId", default_executor_hardware_tier
            )
            master_hardware_tier_id = on_demand_spark_cluster_properties.get(
                "masterHardwareTierId", default_master_hardware_tier
            )
            executor_count = int(
                on_demand_spark_cluster_properties.get("executorCount", 1)
            )
            compute_environment_id = on_demand_spark_cluster_properties.get(
                "computeEnvironmentId"
            )
            executor_storage = int(
                on_demand_spark_cluster_properties.get("executorStorageMB", 0)
            )
            spark_cluster_properties = {
                "computeEnvironmentId": compute_environment_id,
                "executorCount": executor_count,
                "executorHardwareTierId": executor_hardware_tier_id,
                "executorStorageMB": executor_storage,
                "masterHardwareTierId": master_hardware_tier_id,
            }

        resolved_hardware_tier_id = (
            hardware_tier_id or self.get_hardware_tier_id_from_name(hardware_tier_name)
        )
        url = self._routes.job_start()
        payload = {
            "projectId": self.project_id,
            "commandToRun": command,
            "commitId": commit_id,
            "overrideHardwareTierId": resolved_hardware_tier_id,
            "onDemandSparkClusterProperties": spark_cluster_properties,
            "computeClusterProperties": validated_compute_cluster_properties,
            "environmentId": environment_id,
            "externalVolumeMounts": external_volume_mounts,
            "title": title,
        }
        try:
            response = self.request_manager.post(url, json=payload)
            return response.json()
        except exceptions.ReloginRequiredException:
            self._logger.info(
                f" You need to log in to the Domino UI to start the job. Please do it at {self._routes.host}/relogin?redirectPath=/"
            )

    def job_stop(self, job_id: str, commit_results: bool = True):
        """
        Stops the Job with given job_id
        :param job_id: The job identifier
        :param commit_results: Boolean representing to commit results to file
        """
        url = self._routes.job_stop()
        request = {
            "projectId": self.project_id,
            "jobId": job_id,
            "commitResults": commit_results,
        }
        response = self.request_manager.post(url, json=request)
        return response

    def jobs_list(self,
                  project_id: str,
                  order_by: str = "number",
                  sort_by: str = "desc",
                  page_size: Optional[int] = None,
                  page_no: int = 1,
                  show_archived: str = "false",
                  status: str = "all",
                  tag: Optional[str] = None):
        """
        Lists job history for a given project_id
        :param project_id: The project to query
        :param order_by: Field on which sort has to be applied– e.g. "title" (default "number")
        :param sort_by: Sort "desc" (default) or "asc"
        :param page_size: The number of jobs to return (default: 3)
        :param page_no: Page number to fetch (default: 1).
        :param show_archived: Show archived jobs in results (default: false)
        :param status: Status of jobs to fetch– e.g. "completed" (default: "all")
        :param tag: Optional tag filter
        :return: The details
        """
        url = self._routes.jobs_list(project_id, order_by, sort_by, page_size, page_no, show_archived, status, tag)
        return self._get(url)


    def job_status(self, job_id: str) -> dict:
        """
        Gets the status of job with given job_id
        :param job_id: The job identifier
        :return: The details
        """
        return self.request_manager.get(self._routes.job_status(job_id)).json()

    def job_restart(
            self,
            job_id:str,
            should_use_original_input_commit: bool = True
        ):
        """
        Restarts a previous job
        :param job_id:                              ID of the original job that should be restarted
        :param should_use_original_input_commit:    Should the new job run use the original code, or the current version?
        """
        url = self._routes.job_restart()
        request = {
            "jobId": job_id,
            "shouldUseOriginalInputCommit": should_use_original_input_commit
        }
        response = self.request_manager.post(url, json=request)
        return response

    def job_runtime_execution_details(self, job_id: str) -> dict:
        """
        Gets the runtime execution details of job with given job_id
        :param job_id: The job identifier
        :return: The details
        """
        return self.request_manager.get(
            self._routes.job_runtime_execution_details(job_id)
        ).json()

    def job_start_blocking(
        self,
        poll_freq: int = 5,
        max_poll_time: int = 6000,
        ignore_exceptions: Tuple = (requests.exceptions.RequestException,),
        **kwargs,
    ) -> dict:
        """
        Starts a job in a blocking loop, periodically polling for status of job
        is complete. Will ignore intermediate request exception.
        :param poll_freq: int (Optional) polling frequency interval in seconds
        :param max_poll_time: int (Optional) max poll time in seconds
        :param ignore_exceptions: tuple (Optional) a tuple of exceptions that can be ignored
        :param kwargs: Additional arguments to be passed to job_start
        :return:
        """

        def get_job_status(job_identifier):
            status = self.job_status(job_identifier)
            self.log.info(
                f"Polling Job: {job_identifier} status is completed: {status['statuses']['isCompleted']}"
            )
            if status["statuses"]["executionStatus"] == "Failed":
                raise exceptions.RunFailedException(
                    f"Run Status Failed for run: {job_identifier}"
                )
            return status

        job = self.job_start(**kwargs)
        job_id = job["id"]
        job_status = polling2.poll(
            target=lambda: get_job_status(job_id),
            check_success=lambda status: status["statuses"]["isCompleted"],
            ignore_exceptions=ignore_exceptions,
            timeout=max_poll_time,
            step=poll_freq,
            log_error=self.log.level,
        )
        stdout_msg = self.get_run_log(runId=job_id, includeSetupLog=False)
        self.process_log(stdout_msg)
        return job_status

    def files_list(self, commitId, path="/"):
        url = self._routes.files_list(commitId, path)
        return self._get(url)

    def files_upload(self, path, file):
        if not path.startswith("/"):
            path = "/" + path
        url = self._routes.files_upload(path)
        return self._put_file(url, file)

    def blobs_get(self, key):
        """
        Deprecated. Use blobs_get_v2(path, commit_id, project_id) instead
        :param key: blob key
        :return: blob content
        """
        message = "blobs_get is deprecated and will soon be removed. Please migrate to blobs_get_v2 and adjust the " \
                  "input parameters accordingly "
        warnings.warn(message, DeprecationWarning)
        self._validate_blob_key(key)
        url = self._routes.blobs_get(key)
        return self.request_manager.get_raw(url)

    def blobs_get_v2(self, path, commit_id, project_id):
        self._validate_blob_path(path)
        url = self._routes.blobs_get_v2(path, commit_id, project_id)
        return self.request_manager.get_raw(url)

    def fork_project(self, target_name):
        url = self._routes.fork_project(self.project_id)
        request = {"name": target_name}
        response = self.request_manager.post(url, json=request)
        return response

    def endpoint_state(self):
        url = self._routes.endpoint_state()
        return self._get(url)

    def endpoint_unpublish(self):
        url = self._routes.endpoint()
        response = self.request_manager.delete(url)
        return response

    def endpoint_publish(self, file, function, commitId):
        url = self._routes.endpoint_publish()

        request = {
            "commitId": commitId,
            "bindingDefinition": {"file": file, "function": function},
        }

        response = self.request_manager.post(url, json=request)
        return response

    def deployment_version(self):
        url = self._routes.deployment_version()
        return self._get(url)

    def project_create_v4(
        self,
        project_name: str,
        owner_id: Optional[str] = None,
        owner_username: Optional[str] = None,
        description: Optional[str] = "",
        collaborators: Optional[list] = None,
        tags: Optional[list] = None,
        billing_tag: Optional[str] = None,
        visibility: Optional[ProjectVisibility] = ProjectVisibility.PUBLIC,
    ):
        owner = (
            owner_id if owner_id else self.get_user_id(owner_username) if owner_username
            else self.get_user_id(self._owner_username)
        )
        data = {
            "name": project_name,
            "visibility": visibility.value,
            "ownerId": owner,
            "description": description,
            "collaborators": collaborators if collaborators is not None else [],
            "tags": {"tagNames": tags if tags is not None else []},
        }

        if billing_tag:
            data.update({"billingTag": {"tag": billing_tag}})

        url = self._routes.project_v4()
        payload = json.dumps(data)
        response = self.request_manager.post(url, data=payload, headers={'Content-Type': 'application/json'})
        return response.json()

    def project_create(self, project_name, owner_username=None):
        url = self._routes.project_create()
        data = {"projectName": project_name, "ownerOverrideUsername": owner_username}
        response = self.request_manager.post(
            url, data=data, headers=self._csrf_no_check_header
        )
        return response

    def project_archive(self, project_name):
        """Delete the project with the given name"""
        all_owned_projects = self.projects_list(show_completed=True)
        for p in all_owned_projects:
            if p["name"] == project_name:
                url = self._routes.project_v4(project_id=p["id"])
                self.request_manager.delete(url)
                break
        else:
            raise exceptions.ProjectNotFoundException(project_name)

    def projects_list(self, relationship="Owned", show_completed=False):
        url = self._routes.projects_list()

        valid_relationships = ["Owned", "Collaborating", "Suggesting", "Popular"]
        if relationship not in valid_relationships:
            raise ValueError(
                f"relationship must be one of {valid_relationships}: {relationship}"
            )

        query = {
            "relationship": relationship,
            "showCompleted": str(show_completed).lower(),
        }
        response = self.request_manager.get(url, params=query)
        return response.json()

    def tags_list(self, project_id=None):
        project_id = project_id if project_id else self.project_id
        url = self._routes.project_v4(project_id)
        return self._get(url)["tags"]

    def tag_details(self, tag_id):
        url = self._routes.tag_details(tag_id)
        return self._get(url)

    def tags_add(self, tags: list, project_id=None):
        project_id = project_id if project_id else self.project_id
        js_body = {"tagNames": tags}
        url = self._routes.tags_add(project_id)
        response = self.request_manager.post(url, json=js_body)
        return response

    def tag_get_id(self, tag_name, project_id=None):
        tags = self.tags_list(project_id)
        for tag in tags:
            if tag_name == tag["name"]:
                return tag["id"]
        logging.warning(f"tag with name {tag_name} not found for this project")
        return

    def tags_remove(self, tag_name, project_id=None):
        project_id = project_id if project_id else self.project_id
        tag_id = self.tag_get_id(tag_name, project_id)

        if tag_id:
            url = self._routes.tags_remove(project_id, tag_id)
            response = self.request_manager.delete(url)
            return response
        return

    def collaborators_get(self):
        url = self._routes.collaborators_get()
        return self._get(url)

    def get_user_id(self, username_or_email):
        """
        Return the user ID of the the user account matching the given search criteria.

        Args:
            username_or_email (str): the username or email account of a user

        Return:
           user_id (str): the matching user ID, or None if not found
        """
        url = self._routes.users_get()
        response = self.request_manager.get(url)
        users = response.json()

        user_id = None
        for user in users:
            if username_or_email in (user.get("userName"), user.get("email")):
                user_id = user["id"]
                break
        return user_id

    def collaborators_add(self, username_or_email, message=""):
        user_id = self.get_user_id(username_or_email)

        if user_id is None:
            raise exceptions.UserNotFoundException(
                f"Could not add collaborator matching {username_or_email}"
            )

        url = self._routes.collaborators_add(self.project_id)
        request = {"collaboratorId": user_id, "projectRole": "Contributor"}

        response = self.request_manager.post(url, json=request)
        return response

    def collaborators_remove(self, username_or_email):
        user_id = self.get_user_id(username_or_email)

        if user_id is None:
            raise exceptions.UserNotFoundException(
                f"Could not remove collaborator matching {username_or_email}"
            )

        url = self._routes.collaborators_remove(self.project_id, user_id)

        response = self.request_manager.delete(url)
        return response

    # App functions
    def app_publish(self, unpublishRunningApps=True, hardwareTierId=None):
        if unpublishRunningApps:
            self.app_unpublish()
        app_id = self._app_id
        if app_id is None:
            # No App Exists creating one
            app_id = self.__app_create(hardware_tier_id=hardwareTierId)
        url = self._routes.app_start(app_id)
        request = {"hardwareTierId": hardwareTierId}
        response = self.request_manager.post(url, json=request)
        return response

    def app_unpublish(self):
        app_id = self._app_id
        if app_id is None:
            return
        status = self.__app_get_status(app_id)
        self.log.debug(f"App {app_id} status={status}")
        if status and status != "Stopped" and status != "Failed":
            url = self._routes.app_stop(app_id)
            response = self.request_manager.post(url)
            return response

    def __app_get_status(self, id) -> Optional[str]:
        app_id = self._app_id
        if app_id is None:
            return None
        url = self._routes.app_get(app_id)
        response = self.request_manager.get(url).json()
        return response.get("status", None)

    def __app_create(self, name: str = "", hardware_tier_id: str = None) -> str:
        """
        Private method to create app

        :param name: Optional field to set title of app
        :param hardware_tier_id: Optional field to override hardware tier
        :return: Id of newly created App (Un-Published)
        """
        url = self._routes.app_create()
        request_payload = {
            "modelProductType": "APP",
            "projectId": self.project_id,
            "name": name,
            "owner": "",
            "created": time.time_ns(),
            "lastUpdated": time.time_ns(),
            "status": "",
            "media": [],
            "openUrl": "",
            "tags": [],
            "stats": {"usageCount": 0},
            "appExtension": {"appType": ""},
            "id": "000000000000000000000000",
            "permissionsData": {
                "visibility": "GRANT_BASED",
                "accessRequestStatuses": {},
                "pendingInvitations": [],
                "discoverable": True,
                "appAccessStatus": "ALLOWED",
            },
        }
        r = self.request_manager.post(url, json=request_payload)
        response = r.json()
        key = "id"
        if key in response.keys():
            app_id = response[key]
        else:
            raise Exception("Cannot create app")
        return app_id

    # Environment functions
    def environments_list(self):
        url = self._routes.environments_list()
        return self._get(url)

    def get_default_environment(self) -> Dict[str, Any]:
        url = self._routes.environment_default_get()
        response = self.request_manager.get(url).json()
        return response

    def get_environment(self, environment_id: str):
        url = self._routes.environment_get(environment_id)
        response = self.request_manager.get(url).json()
        return response

    def archive_environment(self, environment_id: str) -> None:
        url = self._routes.environment_get(environment_id)
        self.request_manager.delete(url)


    def create_environment(
            self,
            name: str,
            visibility: str,
            dockerfile_instructions: str = "",
            environment_variables: Optional[List[Dict[str, Any]]] = None,
            base_image: str = "",
            post_run_script: str = "",
            post_setup_script: str = "",
            pre_run_script: str = "",
            pre_setup_script: str = "",
            skip_cache: bool = False,
            summary: str = "",
            supported_clusters: Optional[List[str]] = None,
            tags: Optional[List[str]] = None,
            use_vpn: bool = False,
            workspace_tools: Optional[List[Dict[str, Any]]] = None,
            add_base_dependencies: bool = True,
            description: str = "",
            is_restricted: bool = False,
            organization_owner_id: Optional[str] = None,
    ) -> dict:
        """
        Create a new Domino compute environment.
        
        Args:
            name: Name of the compute environment
            visibility: Visibility level ("Private" or "Global")
            dockerfile_instructions: Dockerfile instructions to customize the environment
            environment_variables: List of environment variables as dicts with 'name' and 'value' keys
            base_image: Base Docker image (if empty, uses default environment image)
            post_run_script: Script to run after the main execution
            post_setup_script: Script to run after environment setup
            pre_run_script: Script to run before the main execution
            pre_setup_script: Script to run before environment setup
            skip_cache: Whether to skip Docker layer caching
            summary: Brief description of the environment
            supported_clusters: List of supported distributed compute cluster types
            tags: List of tags for the environment
            use_vpn: Whether this environment should be able to use VPN
            workspace_tools: Configuration of IDEs that can be used in workspaces
            add_base_dependencies: Whether to include base dependencies
            description: Detailed description
            is_restricted: Whether the environment is restricted
            organization_owner_id: ID of an organization that will own the environment
            
        Returns:
            dict: Created environment details
        """

        if not base_image:
            default_environment = self.get_default_environment()
            base_image = default_environment["latestRevisionDetails"]["dockerImage"]

        # Handle None values for optional parameters
        if environment_variables is None:
            environment_variables = []
        if supported_clusters is None:
            supported_clusters = []
        if tags is None:
            tags = []
        if workspace_tools is None:
            workspace_tools = []

        data = {
            "dockerfileInstructions": dockerfile_instructions,
            "environmentVariables": environment_variables,
            "image": base_image,
            "postRunScript": post_run_script,
            "postSetupScript": post_setup_script,
            "preRunScript": pre_run_script,
            "preSetupScript": pre_setup_script,
            "skipCache": skip_cache,
            "summary": summary,
            "supportedClusters": supported_clusters,
            "tags": tags,
            "useVpn": use_vpn,
            "workspaceTools": workspace_tools,
            "addBaseDependencies": add_base_dependencies,
            "description": description,
            "isRestricted": is_restricted,
            "name": name,
            "visibility": visibility
        }

        if organization_owner_id:
            data.update({
                "orgOwnerId": organization_owner_id
            })

        url = self._routes.environment_create()
        payload = json.dumps(data)
        response = self.request_manager.post(url, data=payload, headers={"Content-Type": "application/json"})
        return response.json()


    def create_environment_revision(
            self,
            environment_id: str,
            dockerfile_instructions: str = "",
            environment_variables: Optional[List[Dict[str, Any]]] = None,
            base_image: Optional[str] = None,
            post_run_script: str = "",
            post_setup_script: str = "",
            pre_run_script: str = "",
            pre_setup_script: str = "",
            skip_cache: bool = False,
            summary: str = "",
            supported_clusters: Optional[List[str]] = None,
            tags: Optional[List[str]] = None,
            use_vpn: bool = False,
            workspace_tools: Optional[List[Dict[str, Any]]] = None,
        ) -> dict:
        """
        Create a new revision of an existing Domino environment.
        
        Args:
            environment_id: ID of the environment for which to create a revision
            dockerfile_instructions: Dockerfile instructions to customize the environment
            environment_variables: List of environment variables as dicts with 'name' and 'value' keys
            base_image: Base Docker image (if None, uses default environment image)
            post_run_script: Script to run after the main execution
            post_setup_script: Script to run after environment setup
            pre_run_script: Script to run before the main execution
            pre_setup_script: Script to run before environment setup
            skip_cache: Whether to skip Docker layer caching
            summary: Brief description of the environment revision
            supported_clusters: List of supported cluster types
            tags: List of tags for the environment
            use_vpn: Whether to use VPN for this environment
            workspace_tools: List of workspace tools configuration
            
        Returns:
            dict: Response content from the API
        """

        if not base_image:
            default_environment = self.get_default_environment()
            base_image = default_environment["latestRevisionDetails"]["dockerImage"]

        # Handle None values for optional parameters
        if environment_variables is None:
            environment_variables = []
        if supported_clusters is None:
            supported_clusters = []
        if tags is None:
            tags = []
        if workspace_tools is None:
            workspace_tools = []

        data = {
            "dockerfileInstructions": dockerfile_instructions,
            "environmentVariables": environment_variables,
            "image": base_image,
            "postRunScript": post_run_script,
            "postSetupScript": post_setup_script,
            "preRunScript": pre_run_script,
            "preSetupScript": pre_setup_script,
            "skipCache": skip_cache,
            "summary": summary,
            "supportedClusters": supported_clusters,
            "tags": tags,
            "useVpn": use_vpn,
            "workspaceTools": workspace_tools
        }

        url=self._routes.revision_create(environment_id)
        payload=json.dumps(data)
        response = self.request_manager.post(url, data=payload, headers={"Content-Type": "application/json"})
        return response.json()


    def restrict_environment_revision(
            self,
            environment_id: str,
            revision_id: str
       ) -> None:
        """
        Restrict an environment revision.
        """

        data = {
            "isRestricted": True
        }

        url=self._routes.revision_patch(environment_id, revision_id)
        payload=json.dumps(data)
        self.request_manager.patch(url, data=payload, headers={"Content-Type": "application/json"})


    # Model Manager functions
    def models_list(self):
        url = self._routes.models_list()
        return self._get(url)

    def model_publish(
        self, file, function, environment_id, name, description, files_to_exclude=None
    ):
        if files_to_exclude is None:
            files_to_exclude = []
        url = self._routes.model_publish()

        request = {
            "name": name,
            "description": description,
            "projectId": self.project_id,
            "file": file,
            "function": function,
            "excludeFiles": files_to_exclude,
            "environmentId": environment_id,
        }

        response = self.request_manager.post(url, json=request)
        return response.json()

    def model_versions_get(self, model_id):
        url = self._routes.model_versions_get(model_id)
        return self._get(url)

    def model_version_publish(
        self,
        model_id,
        file,
        function,
        environment_id,
        description,
        files_to_exclude=None,
    ):
        if files_to_exclude is None:
            files_to_exclude = []
        url = self._routes.model_version_publish(model_id)

        request = {
            "description": description,
            "projectId": self.project_id,
            "file": file,
            "function": function,
            "excludeFiles": files_to_exclude,
            "environmentId": environment_id,
        }

        response = self.request_manager.post(url, json=request)
        return response.json()

    # Dataset Functions
    def datasets_list(self, project_id=None):
        url = self._routes.datasets_list(project_id)
        return self._get(url)

    def datasets_ids(self, project_id) -> list:
        dataset_list = self.datasets_list(project_id)
        dataset_ids = [dataset_id["datasetId"] for dataset_id in dataset_list]
        return dataset_ids

    def datasets_names(self, project_id) -> list:
        dataset_list = self.datasets_list(project_id)
        dataset_names = [dataset_id["datasetName"] for dataset_id in dataset_list]
        return dataset_names

    def datasets_create(self, dataset_name, dataset_description):
        if dataset_name in self.datasets_names(self.project_id):
            raise exceptions.DatasetExistsException("Dataset Name must be Unique")

        url = self._routes.datasets_create()
        request = {
            "datasetName": dataset_name,
            "description": dataset_description,
            "projectId": self.project_id,
        }
        response = self.request_manager.post(url, json=request)
        return response.json()

    def datasets_details(self, dataset_id):
        url = self._routes.datasets_details(dataset_id)
        return self._get(url)

    def datasets_update_details(
        self, dataset_id, dataset_name=None, dataset_description=None
    ):
        url = self._routes.datasets_details(dataset_id)
        request = {}
        if dataset_name:
            if dataset_name in self.datasets_names(self.project_id):
                raise exceptions.DatasetExistsException("Dataset Name must be Unique")
            else:
                request.update({"datasetName": dataset_name})
        if dataset_description:
            request.update({"description": dataset_description})

        self.request_manager.patch(url, json=request)

        return self._get(url)

    def _dataset_remove(self, dataset_id):
        if dataset_id in self.datasets_ids(self.project_id):
            raise exceptions.DatasetNotFoundException(
                f"Dataset with id {dataset_id} does not exist"
            )
        url = self._routes.datasets_details(dataset_id)
        response = self.request_manager.delete(url)
        return response

    def datasets_remove(self, dataset_ids: list):
        for dataset_id in dataset_ids:
            url = self._routes.datasets_details(dataset_id)
            self.request_manager.delete(url)

    def datasets_upload_files(
        self,
        dataset_id: str,
        local_path_to_file_or_directory: str,
        file_upload_setting: str = None,
        max_workers: int = None,
        target_chunk_size: int = None,
        target_relative_path: str = None
    ) -> str:
        """Upload file to dataset with multithreaded support.

        Note: in the case of a KeyboardInterrupt, it may take a few seconds for all the threads to stop and for the
        upload session to be properly cancelled.

        Args:
            dataset_id: id of dataset whose rw snapshot the file will be uploaded to
            local_path_to_file_or_directory: path to file or directory in local machine
            file_upload_setting: setting to resolve naming conflict, one of Rename, Overwrite, or Ignore (default).
                - Rename: renames the conflicting file being uploaded by appending "_n" where n is an integer.
                - Overwrite: overrides the existing conflicting files.
                - Ignore (default): skips the upload for any conflicting file.
            max_workers: max amount of threads (default: 10)
            target_chunk_size: max chunk size for multipart upload (default: 8MB)
            target_relative_path: path in the dataset to upload the file or directory to
        Returns local path to uploaded file
        """
        if file_upload_setting is None or file_upload_setting == "Ignore":
            text = "Ignore setting selected - any file with naming conflict will not be uploaded."
        elif file_upload_setting == "Overwrite":
            text = "Overwrite setting selected - note that any existing file with naming conflict " \
                   "will be overridden."
        elif file_upload_setting == "Rename":
            text = "Rename setting selected - note that naming conflicts will be resolved by appending " \
                   "an increasing integer at the end of the uploaded files. In case of a directory with " \
                   "numerous conflicts, this will cause severe file proliferation."
        else:
            raise ValueError(f"input file_upload_setting {file_upload_setting} not allowed. Please use "
                             f"`Overwrite`, `Rename`, or `Ignore` only.")
        self.log.warning(text)

        with datasets.Uploader(
            csrf_no_check_header=self._csrf_no_check_header,
            dataset_id=dataset_id,
            local_path_to_file_or_directory=local_path_to_file_or_directory,
            log=self.log,
            request_manager=self.request_manager,
            routes=self._routes,

            file_upload_setting=file_upload_setting,
            max_workers=max_workers,
            target_chunk_size=target_chunk_size,
            target_relative_path=target_relative_path
        ) as uploader:
            path = uploader.upload()
            self.log.info(f"Uploading chunks for file or directory `{path}` to dataset {dataset_id} completed. "
                          f"Now attempting to end upload session.")
            return path

    def model_version_export(
        self,
        model_id,
        model_version_id,
        registry_host,
        registry_username,
        registry_password,
        repository_name,
        image_tag,
    ):
        self.requires_at_least("4.1.0")

        url = self._routes.model_version_export(model_id, model_version_id)

        request = {
            "registryUrl": registry_host,
            "username": registry_username,
            "password": registry_password,
            "repository": repository_name,
            "tag": image_tag,
        }

        response = self.request_manager.post(url, json=request)
        return response.json()

    def model_version_sagemaker_export(
        self,
        model_id,
        model_version_id,
        registry_host,
        registry_username,
        registry_password,
        repository_name,
        image_tag,
    ):
        self.requires_at_least("4.2.0")

        url = self._routes.model_version_sagemaker_export(model_id, model_version_id)

        request = {
            "registryUrl": registry_host,
            "username": registry_username,
            "password": registry_password,
            "repository": repository_name,
            "tag": image_tag,
        }

        response = self.request_manager.post(url, json=request)
        return response.json()

    def model_version_export_status(self, model_export_id):
        self.requires_at_least("4.1.0")

        url = self._routes.model_version_export_status(model_export_id)
        return self._get(url)

    def model_version_export_logs(self, model_export_id):
        self.requires_at_least("4.1.0")

        url = self._routes.model_version_export_logs(model_export_id)
        return self._get(url)

    # Hardware Tier Functions
    def hardware_tiers_list(self):
        url = self._routes.hardware_tiers_list(self.project_id)
        return self._get(url)

    def get_hardware_tier_id_from_name(self, hardware_tier_name: str):
        for hardware_tier in self.hardware_tiers_list():
            if hardware_tier_name == hardware_tier["hardwareTier"]["name"]:
                return hardware_tier["hardwareTier"]["id"]
        return None

    def process_log(self, log):
        """
        spool out and replay entire log using bs4 to strip the HTML tags
            - html.parser since it's batteries included
        """
        for line in log.splitlines():
            if line and line != ".":
                text = BeautifulSoup(line, "html.parser").text

                if "text-danger" in line:
                    self.log.warning(text)
                else:
                    self.log.info(text)

    def budget_defaults_list(self) -> list:
        """
        Get a list of the available default budgets with the assigned (if any) limits
        Requires Admin permission

        :return: Returns a list of default budgets with details
        """
        url = self._routes.budgets_default()
        return self.request_manager.get(url).json()

    def budget_defaults_update(self, budget_label: BudgetLabel, budget_limit: float) -> dict:
        """
        Update default budgets limits (or quota) by BudgetLabel
        Requires Admin permission

        :param budget_label: label of budget to be updated ex: `BillingTag`, `Organization`
        :param budget_limit: new budget quota to assign to default label

        :return: Returns the updated budget with the newly assigned limit.
        """
        url = self._routes.budgets_default(budget_label.value)
        updated_budget = {"budgetLabel": budget_label.value, "budgetType": "Default", "limit": budget_limit,
                          "window": "monthly"}
        data = json.dumps(updated_budget)
        return self.request_manager.put(url, data=data, headers={"Content-Type": "application/json"}).json()

    def budget_overrides_list(self):
        """
        Get a list of the available budgets overrides with the assigned limits.
        Requires Admin permission

        :return: Returns a list of budgets overrides with details
        """
        url = self._routes.budget_overrides()
        return self.request_manager.get(url).json()

    def budget_override_create(self, budget_label: BudgetLabel, budget_id: str, budget_limit: float) -> dict:
        """
        Create Budget overrides based on BudgetLabels, ie BillingTags, Organization, or Projects
        the object id is used as budget ids
        Requires Admin roles

        :param budget_label: label of budget to be updated
        :param budget_id: id of project or organization to be used as new budget override id.
        :param budget_limit: budget quota to assign to override

        :return: Returns the newly created budget override
        """
        url = self._routes.budget_overrides()
        new_budget: dict = self._generate_budget(budget_label, budget_id, budget_limit)
        data = json.dumps(new_budget)
        return self.request_manager.post(url, data=data, headers={"Content-Type": "application/json"}).json()

    def budget_override_update(self, budget_label: BudgetLabel, budget_id: str, budget_limit: float) -> dict:
        """
        Update Budget overrides based on BudgetLabel and budget id
        Requires Admin roles

        :param budget_label: label of budget to be updated
        :param budget_id: id of budget override to be updated.
        :param budget_limit: new budget quota to assign to override

        :return: Returns the updated budget override
        """
        url = self._routes.budget_overrides(budget_id)
        new_budget: dict = self._generate_budget(budget_label, budget_id, budget_limit)
        data = json.dumps(new_budget)
        return self.request_manager.put(url, data=data, headers={"Content-Type": "application/json"}).json()

    def budget_override_delete(self, budget_id: str) -> list:
        """
        Delete an existing budget override
        Requires Admin roles

        :param budget_id: id of budget override to be deleted.

        :return: Returns list of messages confirming the budget override deletion
        """
        url = self._routes.budget_overrides(budget_id)
        return self.request_manager.delete(url).json()

    def budget_alerts_settings(self) -> dict:
        """
        Get the current budget alerts settings
        Requires Admin permission

        :return: Returns the current budget alert setting
        """
        url = self._routes.budget_settings()
        return self.request_manager.get(url).json()

    def budget_alerts_settings_update(
        self,
        alerts_enabled: Optional[bool] = None,
        notify_org_owner: Optional[bool] = None
    ) -> dict:
        """
        Update the current budget alerts settings to enable/disable budget notifications
        and whether to notify org owners on projects notifications
        Requires Admin permission

        :param alerts_enabled: whether to enable or disable notifications.
        :param notify_org_owner: whether to notify organizations owners on projects reaching threshold.

        :return: Returns the updated budget alert setting
        """
        current_settings = self.budget_alerts_settings()

        optional_fields = {
            "alertsEnabled": alerts_enabled,
            "notifyOrgOwner": notify_org_owner
        }

        updated_settings = self._update_if_set(current_settings, optional_fields)

        payload = json.dumps(updated_settings)
        url = self._routes.budget_settings()
        return self.request_manager.put(url, data=payload).json()

    def budget_alerts_targets_update(self, targets: dict[BudgetLabel, list]) -> dict:
        """
        Update the current budget alerts settings with additional email targets per budget label
        Requires Admin permission

        :param targets: dictionary of budget labels and list of email addresses

        :return: Returns the updated budget alert setting
        """

        budget_settings = self.budget_alerts_settings()

        current_targets = budget_settings["alertTargets"]

        updated_targets = []

        for target in current_targets:
            if target["label"] in targets:
                updated_targets.append({"label": target["label"], "emails": targets[target["label"]]})
            else:
                updated_targets.append(target)

        budget_settings["alertTargets"] = updated_targets

        payload = json.dumps(budget_settings)
        url = self._routes.budget_settings()
        return self.request_manager.put(url, data=payload).json()

    def billing_tags_list_active(self) -> dict:
        """
        Get a list of active billing tags
        Requires Admin permission

        :return: Returns a dictionary containing the list active billing tags
        """
        self.requires_at_least("5.11.0")
        url = self._routes.billing_tags()
        return self.request_manager.get(url).json()

    def billing_tags_create(self, tags_list: list) -> dict:
        """
        Create a list of active billing tags
        Requires Admin permission

        :param tags_list: list of billing tags names to be created

        :return: Returns a dictionary containing the list newly created billing tags
        """
        self.requires_at_least("5.11.0")
        url = self._routes.billing_tags()
        payload = json.dumps({"billingTags": tags_list})
        return self.request_manager.post(url, data=payload, headers={"Content-Type": "application/json"}).json()

    def active_billing_tag_by_name(self, name: str) -> dict:
        """
        Get detailed info on active or archived billing tag
        Requires Admin permission

        :param name: name of existing billing tag

        :return: Returns a dictionary containing the list newly created billing tags
        """
        url = self._routes.billing_tags(name)
        return self.request_manager.get(url).json()

    def billing_tag_archive(self, name: str) -> dict:
        """
        Archive an active billing tag
        Requires Admin permission

        :param name: name of existing billing tag to archive

        :return: Returns a dictionary containing the updated billing tag
        """
        url = self._routes.billing_tags(name)
        return self.request_manager.delete(url).json()

    def billing_tag_settings(self) -> dict:
        """
        Get the current billing tag settings
        Requires Admin permission

        :return: Returns the current billing tag settings
        """
        url = self._routes.billing_tags_settings()
        return self.request_manager.get(url).json()

    def billing_tag_settings_mode(self) -> dict:
        """
        Get the current billing tag settings mode
        Requires Admin permission

        :return: Returns the current billing tag settings mode
        """
        url = self._routes.billing_tags_settings(mode_only=True)
        return self.request_manager.get(url).json()

    def billing_tag_settings_mode_update(self, mode: BillingTagSettingMode) -> dict[str, BillingTagSettingMode]:
        """
        Update the current billing tag settings mode
        Requires Admin permission

        :param mode: new mode to set the billing tag settings (see BillingTagSettingMode)

        :return: Returns the updated billing tag settings mode
        """
        url = self._routes.billing_tags_settings(mode_only=True)
        payload = json.dumps({"mode": mode.value})
        return self.request_manager.put(url, data=payload, headers={"Content-Type": "application/json"}).json()

    def project_billing_tag(self, project_id: Optional[str] = None) -> Optional[dict]:
        """
        Get a billing tag assigned to a particular project by project id
        Requires Admin permission

        :param project_id: id of the project to find assigned billing tag

        :return: Returns the billing tag if assigned or None
        """
        url = self._routes.project_billing_tag(project_id if project_id else self.project_id)
        return self.request_manager.get(url).json()

    def project_billing_tag_update(self, billing_tag: str, project_id: Optional[str] = None) -> dict:
        """
        Update project's billing tag with new billing tag.
        Requires Admin permission

        :param billing_tag: billing tag to assign to a project
        :param project_id: id of the project to assign a billing tag

        :return: Returns the project details including the new billing tag
        """
        url = self._routes.project_billing_tag(project_id if project_id else self.project_id)
        data = {
            "tag": billing_tag
        }
        return self.request_manager.post(url, data=json.dumps(data)).json()

    def project_billing_tag_reset(self, project_id: Optional[str] = None) -> dict:
        """
        Remove a billing tag from a specified project
        Requires Admin permission

        :param project_id: id of the project to reset billing tag field

        :return: Returns the project details
        """
        url = self._routes.project_billing_tag(project_id if project_id else self.project_id)
        return self.request_manager.delete(url).json()

    def projects_by_billing_tag(
        self,
        billing_tag: Optional[str] = None,
        offset: Optional[int] = 0,
        page_size: Optional[int] = 10,
        name_filter: Optional[str] = None,
        sort_by: Optional[str] = None,
        sort_order: Optional[str] = None,
        missing_tag_only: bool = False,
    ) -> dict:
        """
        Remove a billing tag from a specified project
        Requires Admin permission

        :param billing_tag: billing tag string to filter projects by
        :param offset: The index of the start of the page, where checkpointProjectId is index 0.
                        If the offset is negative the project it points to will be the end of the page.
        :param page_size: The number of record to return per page.
        :param name_filter: matches projects by name substring
        :param sort_by: (Optional) field to sort the projects on
        :param sort_order: (Optional) Whether to sort in asc or desc order
        :param missing_tag_only: (Optional) determine whether to only return projects with missing tag

        :return: Returns a dictionary containing a page with the projects matching the query based on the page_size,
                 and to total count of projects aching the query
        """

        parameters = {
            "offset": offset,
            "pageSize": page_size,
            "missingBillingTagOnly": str(missing_tag_only).lower()
        }

        optional_params = {
            "billingTag": billing_tag,
            "nameFilter": name_filter,
            "sortBy": sort_by,
            "sortOrder": sort_order,
        }

        updated_params = self._update_if_set(parameters, optional_params)

        url = self._routes.projects_billing_tags()

        return self.request_manager.get(url, params=updated_params).json()

    def project_billing_tag_bulk_update(self, projects_tag: dict[str, str]) -> dict:
        """
        Update project's billing tags in bulk
        Requires Admin permission

        :param projects_tag: dictionary of project_id and billing_tags

        :return: Returns the updated projects detail summary
        """
        value_list = []
        for key, value in projects_tag.items():
            value_list.append(
                {
                    "projectId": key,
                    "billingTag": {
                        "tag": value
                    }
                }
            )

        data = {
            "projectsBillingTags": value_list
        }

        url = self._routes.projects_billing_tags()
        return self.request_manager.post(url, data=json.dumps(data)).json()

    _CUSTOM_METRICS_USE_GEN = True

    def custom_metrics_client(self) -> _CustomMetricsClientBase:
        self.requires_at_least("5.3.1")
        if self._CUSTOM_METRICS_USE_GEN:
            return _CustomMetricsClientGen(self, self._routes)
        else:
            return _CustomMetricsClient(self, self._routes)

    # Validation methods
    def _useable_environments_list(self):
        self.log.debug("Getting list of useable environment")
        useable_environment_list_url = self._routes.useable_environments_list(
            self.project_id
        )
        return self.request_manager.get(useable_environment_list_url).json()[
            "environments"
        ]

    def _validate_environment_id(self, environment_id) -> bool:
        self.log.debug(f"Validating environment id: {environment_id}")
        for environment in self._useable_environments_list():
            if environment_id == environment["id"]:
                return True
        raise exceptions.EnvironmentNotFoundException(
            f"{environment_id} environment not found"
        )

    def _validate_hardware_tier_id(self, hardware_tier_id: str) -> bool:
        self.log.debug(f"Validating hardware tier id: {hardware_tier_id}")
        for hardware_tier in self.hardware_tiers_list():
            if hardware_tier_id == hardware_tier["hardwareTier"]["id"]:
                return True
        raise exceptions.HardwareTierNotFoundException(
            f"{hardware_tier_id} hardware tier Id not found"
        )

    def _validate_hardware_tier_name(self, hardware_tier_name: str) -> bool:
        self.log.debug(f"Validating hardware tier name: {hardware_tier_name}")
        count = 0
        for hardware_tier in self.hardware_tiers_list():
            if hardware_tier_name == hardware_tier["hardwareTier"]["name"]:
                count += 1

        if count == 0:
            raise exceptions.HardwareTierNotFoundException(
                f"{hardware_tier_name} hardware tier name not found"
            )
        elif count > 1:
            raise exceptions.HardwareTierNotFoundException(
                f"{hardware_tier_name} hardware tier name has more than one match"
            )
        return True

    def _validate_commit_id(self, commit_id):
        self.log.debug(f"Validating commit id: {commit_id}")
        for commit in self.commits_list():
            if commit_id == commit:
                return True
        raise exceptions.CommitNotFoundException(f"{commit_id} commit Id not found")

    # Helper methods
    def _get(self, url):
        return self.request_manager.get(url).json()

    def _put_file(self, url, file):
        return self.request_manager.put(url, data=file)

    @staticmethod
    def _validate_blob_key(key):
        regex = re.compile("^\\w{40,40}$")
        if not regex.match(key):
            raise Exception(
                (
                    "Blob key should be 40 alphanumeric characters. "
                    "Perhaps you passed a file path on accident? "
                    "If you have a file path and want to get the "
                    "file, use files_list to get the blob key."
                )
            )

    @staticmethod
    def _validate_blob_path(path):
        """
        Helper method to validate that the path is normalized
        For example: A//B, A/B/, A/./B and A/foo/../B are not normalized/canonical, whereas A/B is.

        Args:
            path: the string of the path to check

        Return:
           None, however, it throws a MalformedInputException if the input is not normalized/canonical.
        """
        normalized_path = os.path.normpath(path)
        if path != normalized_path:
            raise exceptions.MalformedInputException(
                (
                    "Path should be normalized and cannot contain "
                    "'..' or '../'. "
                )
            )

    @staticmethod
    def _validate_information_data_type(info: dict):
        accepted_units = {"GiB": True, "MB": True}

        try:
            unit = info["unit"]
            _ = info["value"]
            accepted_units[unit]
        except KeyError:
            raise Exception(
                "Information value is formatted incorrectly."
                + f" Allowed units: {', '.join(accepted_units.keys())}"
                + " Example: { 'unit': 'GiB', 'value': 5 }"
            )

    @staticmethod
    def _generate_budget(budget_label: BudgetLabel, budget_id: str, budget_limit: float) -> dict:
        return {
            "limit": budget_limit,
            "labelId": budget_id,
            "window": "monthly",
            "budgetLabel": budget_label.value,
            "budgetType": BudgetType.OVERRIDE.value
        }

    @staticmethod
    def _update_if_set(update_dict: dict, new_dict: dict) -> dict:
        for key, value in new_dict.items():
            if value is not None:
                update_dict[key] = value
        return update_dict

    def requires_at_least(self, at_least_version):
        if version.parse(at_least_version) > version.parse(self._version):
            raise Exception(
                f"You need at least version {at_least_version} but your deployment seems to be running {self._version}"
            )
        return True

    # Workaround to get project ID which is needed for some model functions
    @property  # type: ignore # mypy incorrect error silencer
    @functools.lru_cache()
    def project_id(self):
        url = self._routes.find_project_by_owner_name_and_project_name_url()
        key = "id"
        response = self._get(url)
        if key in response.keys():
            return response[key]

    # This will fetch app_id of app in current project
    @property
    def _app_id(self):
        url = self._routes.app_list(self.project_id)
        response = self._get(url)
        if len(response) != 0:
            app = response[0]
        else:
            return None
        key = "id"
        if key in app.keys():
            app_id = app[key]
        else:
            app_id = None
        return app_id

    _csrf_no_check_header = {"Csrf-Token": "nocheck"}
